; This program is made by Daniel Horchner.
; email: Daniel.Horchner@95.Student.WAU.NL
;        dbjh@usa.net

segment code32 use32

%include "raw32.inc"
%define LOADER          0
%include "loader\raw32mod.inc"
extern  DPMImem, VCPImem, XMSmem

global  main

;32-bit data
BUFFERSIZE      equ     64*1024-1
defaultname     db      'MODULE.RX',0
modulename      dd      defaultname
filehandle      dw      0
buffer          dd      0               ; ptr to file buffer
module          dd      0               ; ptr to module in memory
image           dd      0               ; ptr to image in memory (module+hdr)
orgstack        dd      0
                dw      0

msg0            db      'RAW32 module loader - made by Danil Hrchner.',13,10,'$'
msg1            db      'Not enough memory free.','$'
msg2            db      'An error occurred. Press a key to continue.','$'
msg3            db      13,10,'Shutting down...','$'

;32-bit code
main:
        mov edx,msg0
        call dosprint

        mov eax,[himembase]             ; Ensure himembase is aligned on para
        add eax,[code32a]               ;  (needed for relocation code)
        add eax,0fh
        and al,0f0h
        sub eax,[code32a]
        sub eax,[himembase]
        call gethimem

        push dword [lomembase]
        movzx eax,byte [es:80h]         ; psp[80h]=length of command tail
        cmp eax,0
        je .filenamedone                ; No argument, use default filename
        dec eax
        je .filenamedone                ; Only a space -> default filename
        mov ecx,eax                     ; Save length for copying
        call getlomem
        mov [modulename],eax

        mov edi,eax
        mov esi,82h                     ; Name starts after space
        mov es,[data32sel]
        mov ds,[pspsel]
        rep movsb                       ; Copy argument to filename string
        mov ds,[cs:data32sel]
        mov byte [es:edi],0
.filenamedone:

        mov eax,BUFFERSIZE              ; Allocate file buffer
        call getlomem
        jc near notenoughmem
        mov [buffer],eax

        mov byte [v86r_ah],3dh          ; ah=3dh -> Open file using handle
        mov byte [v86r_al],2            ; ah=open access mode; 2=read/write
        mov eax,[modulename]
        add eax,[code32a]               ; eax=linear address of filename
        mov dl,al
        and edx,0fh
        shr eax,4
        mov word [v86r_ds],ax
        mov word [v86r_dx],dx           ; ds:dx=pointer to ASCII-z path name
        mov al,21h
        int RMCALL_VECT
        test word [v86r_flags],1
        jnz near errexit
        mov ax,[v86r_ax]
        mov [filehandle],ax

        mov byte [v86r_ah],42h          ; ah=42h -> Move file pointer
        mov byte [v86r_al],2            ; al=origin of move; 2=eof + offset
        mov word [v86r_bx],ax           ; bx=file handle
        mov word [v86r_cx],0            ; MSW of offset
        mov word [v86r_dx],0            ; LSW of offset
        mov al,21h
        int RMCALL_VECT
        test word [v86r_flags],1
        jnz near errexit
        mov ax,[v86r_dx]
        shl eax,16
        mov ax,[v86r_ax]                ; dx:ax=new file pointer=file size
        call gethimem                   ; eax=file size
        jc near notenoughmem
        mov [module],eax
                                        ; Restore file ptr to start of file
        mov byte [v86r_ah],42h
        mov byte [v86r_al],0            ; 0=beginning of file + offset
        mov ax,[filehandle]
        mov [v86r_bx],ax
        mov word [v86r_cx],0
        mov word [v86r_dx],0
        mov al,21h
        int RMCALL_VECT
        test word [v86r_flags],1
        jnz near errexit
                                        ; Read the module
        mov es,[data32sel]
        mov edi,[module]
        mov ax,[filehandle]
        mov [v86r_bx],ax                ; bx=file handle
        mov word [v86r_cx],BUFFERSIZE   ; cx=number of bytes to read
        mov eax,[buffer]
        add eax,[code32a]
        mov dl,al
        and edx,0fh
        shr eax,4
        mov [v86r_ds],ax
        mov [v86r_dx],dx                ; ds:dx=pointer to read buffer
        mov al,21h
readmodule:
        mov byte [v86r_ah],3fh          ; ah=3fh -> Read from file using hndl
        int RMCALL_VECT
        test word [v86r_flags],1
        jnz near errexit
        movzx ecx,word [v86r_ax]        ; ax=number of bytes actually read
        cmp ecx,0
        je moduleloaded
        mov esi,[buffer]
        rep movsb                       ; Copy buffer to memory for module
        jmp readmodule

moduleloaded:
        mov byte [v86r_ah],3eh          ; 3eh -> Close file using handle
        mov al,21h
        int RMCALL_VECT
        test word [v86r_flags],1
        jnz near errexit

        pop dword [lomembase]           ; Free memory for file buffer
;
        mov ebx,[module]                ; ebx=ptr to module
        mov edi,ebx
        movzx eax,word [ebx+8]          ; [8]=size of header in 16 byte paras
        shl eax,4
        add edi,eax                     ; edi=ptr to image (module - header)
        mov [image],edi

        movzx ecx,word [ebx+6]          ; [6]=number of relocation items
        jecxz relocation_done
        movzx eax,word [ebx+18h]        ; [18h]=offset in module of first
        add ebx,eax                     ;  relocation item
        mov esi,[image]
        add esi,[code32a]
        shr esi,4
do_relocation:                          ; A reloc item contains the seg:off
        movzx edx,word [ebx+2]          ;  addr of a word in image to which
        shl edx,4                       ;  the real mode cs must be added.
        movzx eax,word [ebx]            ;  This loader adds a '32-bit RM seg'
        add edx,eax                     ;  to a _dword_ -> ALWAYS use a
        add [edi+edx],esi               ;  32-bit reg when loading seg names!
        add ebx,4                       ; ebx=ptr to next relocation item
        dec ecx
        jnz do_relocation
relocation_done:
;
        mov ax,[code32sel]              ; Initialize RAW32 vars in module
        mov edx,edi
        add edx,[code32a]
        call make_dsc_alias
        jc near errexit
        mov [edi+code32selOFF],ax
        mov ax,[data32sel]
        call make_dsc_alias
        jc near errexit
        mov [edi+data32selOFF],ax

        mov ax,[zerosel]
        mov [edi+zeroselOFF],ax
        mov ax,[code16sel]
        mov [edi+code16selOFF],ax
        mov ax,[data16sel]
        mov [edi+data16selOFF],ax
        mov ax,[pspsel]
        mov [edi+pspselOFF],ax
        mov ax,[envsel]
        mov [edi+envselOFF],ax

        mov eax,[code16a]
        mov [edi+code16aOFF],eax
        mov eax,[code32a]
        add eax,edi
        mov [edi+code32aOFF],eax
        mov eax,[lomembase]
        sub eax,edi
        mov [edi+lomembaseOFF],eax
        mov eax,[lomemtop]
        sub eax,edi
        mov [edi+lomemtopOFF],eax
        mov eax,[himembase]
        sub eax,edi
        mov [edi+himembaseOFF],eax
        mov eax,[himemtop]
        sub eax,edi
        mov [edi+himemtopOFF],eax
        mov eax,[totalextmem]
        mov [edi+totalextmemOFF],eax
        mov al,[windows]
        mov [edi+windowsOFF],al
        mov al,[returncode]
        mov [edi+returncodeOFF],al
        mov al,[systemtype]
        mov [edi+systemtypeOFF],al

        push edi
        mov ecx,16/4
        mov esi,IRQ0_vect
        add edi,IRQ0_vectOFF
        rep movsd

        mov ecx,16/4
        mov esi,hextbl
        mov edi,hextblOFF
        add edi,[image]
        rep movsd
        pop edi
;
                                        ; Initialize ptrs to RAW32 functions
        mov dword [edi+exitOFF],@exit
        mov word [edi+exitOFF+4],cs

        mov dword [edi+getmemOFF],modulegetmem
        mov word [edi+getmemOFF+4],cs

        mov dword [edi+getlomemOFF],modulegetlomem
        mov word [edi+getlomemOFF+4],cs

        mov dword [edi+gethimemOFF],modulegethimem
        mov word [edi+gethimemOFF+4],cs

        mov dword [edi+lomemsizeOFF],modulelomemsize
        mov word [edi+lomemsizeOFF+4],cs

        mov dword [edi+himemsizeOFF],modulehimemsize
        mov word [edi+himemsizeOFF+4],cs

        mov dword [edi+getdscOFF],modulegetdsc
        mov word [edi+getdscOFF+4],cs

        mov dword [edi+setdscOFF],modulesetdsc
        mov word [edi+setdscOFF+4],cs

        mov dword [edi+getvectOFF],modulegetvect
        mov word [edi+getvectOFF+4],cs

        mov dword [edi+setvectOFF],modulesetvect
        mov word [edi+setvectOFF+4],cs

        mov dword [edi+map_phys_addrOFF],modulemap_phys_addr
        mov word [edi+map_phys_addrOFF+4],cs

        mov dword [edi+dosshellOFF],moduledosshell
        mov word [edi+dosshellOFF+4],cs

        mov dword [edi+dosprintOFF],moduledosprint
        mov word [edi+dosprintOFF+4],cs

        mov dword [edi+putstrOFF],moduleputstr
        mov word [edi+putstrOFF+4],cs
;
        mov bl,RMCALL_VECT+1            ; Module RM int calls must go through
        mov cx,cs                       ;  this routine, because the register
        mov edx,moduleRMint             ;  structure of the module is only a
        call setvect                    ;  copy of the real register struc
        jc near errexit                 ;  that is used for a RM int

        mov dword [orgstack],esp
        mov word [orgstack+4],ss

        push dword [DPMImem]
        push dword [VCPImem]
        push dword [XMSmem]

        mov ax,[edi+data32selOFF]
        mov ss,ax                       ; RAW32 programs assume ss=ds
        sub esp,edi
        mov ds,ax
        mov fs,ax
        mov es,[cs:pspsel]
        mov gs,[cs:zerosel]
        movzx eax,word [cs:edi+code32selOFF]
        push eax                        ; cs must be on stack as dword
        mov ebx,[cs:module]
        movzx eax,word [cs:ebx+14h]     ; [14h]=program entry point
        push eax
        retf                            ; Jump to module

@exit:
        mov ax,[cs:data32sel]
        mov ds,ax
        mov es,ax
        mov fs,ax
        mov gs,[zerosel]
        lss esp,[orgstack]

        mov edi,[image]
        mov al,[edi+returncodeOFF]
        mov [returncode],al
        mov edx,msg3
        call dosprint
        jmp exit                        ; Return to real/V86 mode

;
notenoughmem:
        mov edx,msg1  
        call dosprint
        mov byte [v86r_ah],0
        mov al,16h
        int RMCALL_VECT
        jmp exit                        ; Return to real/V86 mode

;
errexit:
        mov edx,msg2    
        call dosprint
        mov byte [v86r_ah],0            ; ah=0 -> Wait for key and read char
        mov al,16h
        int RMCALL_VECT
        jmp exit                        ; Return to real/V86 mode

;
; Make an alias of a descriptor
; In:
;   ax = selector of a descriptor of which an alias is wanted
;   edx = linear base address of alias descriptor
; Out:
;   CF=0 -> No error
;     ax = selector of alias descriptor
;   CF=1 -> Error
;     ax = ?
;
; Note:
;   This function makes a copy of the descriptor but changes the base address
;   of the descriptor copy to the address specified in edx
;
make_dsc_alias:
        cmp byte [systemtype],3
        je DPMImake_dsc_alias

        push ebx
        push ecx
        push edx
        push esi
        push edi
        push es
        sub esp,6
        sgdt [esp]
        mov ebx,[esp+2]                 ; Get GDT base address
        add esp,6
        mov esi,eax
        and esi,0fff8h                  ; Clear high word, TI (should be 0!)
        mov es,[zerosel]                ;  and RPL -> right offset in GDT
        mov cx,1                        ; Allocate 1 descriptor
        call getdsc
        jc .errexit
        movzx edi,ax
        mov ecx,[es:ebx+esi]            ; Copy descriptor
        mov [es:ebx+edi],ecx
        mov ecx,[es:ebx+esi+4]
        mov [es:ebx+edi+4],ecx

        mov [es:ebx+edi+2],dx           ; Set descriptor's base
        shr edx,16
        mov byte [es:ebx+edi+4],dl
        mov byte [es:ebx+edi+7],dh
        pop es
        pop edi
        pop esi
        pop edx
        pop ecx
        pop ebx
        clc
        ret

.errexit:
        pop es
        pop edi
        pop esi
        pop edx
        pop ecx
        pop ebx
        stc
        ret

DPMImake_dsc_alias:
        push ebx
        push edx
        push edi
        push es
        sub esp,8                       ; Reserve 8 bytes temp space on stack
        mov bx,ax                       ; bx=selector of descriptor to copy
        mov ax,ss
        mov es,ax                       ; es:edi=ptr to buffer to store dsc
        mov edi,esp
        mov ax,0bh                      ; ax=0bh -> Get descriptor
        int 31h
        jc .errexit

        mov cx,1
        call getdsc                     ; Allocate 1 descriptor
        jc .errexit

        mov [esp+2],dx                  ; Set descriptor's base
        shr edx,16
        mov byte [esp+4],dl
        mov byte [esp+7],dh
        mov bx,ax                       ; Set descriptor just allocated
        mov ax,0ch                      ; ax=0ch -> Set descriptor
        int 31h
        jc .errexit
        mov ax,bx                       ; ax=selector of new dsc on return
        add esp,8                       ; Restore stack
        pop es
        pop edi
        pop edx
        pop ebx
        clc
        ret

.errexit:
        add esp,8
        pop es
        pop edi
        pop edx
        pop ebx
        stc
        ret

; Wrapper functions
;
moduleRMint:
        push ecx
        push esi
        push edi
        push ds
        push es
        pushfd
        mov ds,[cs:data32sel]
        mov es,[cs:data32sel]
        mov esi,[image]
        add esi,REGSTRUCOFF
        mov edi,v86r_edi
        mov ecx,25                      ; reg structure is 50 bytes in size
        cld
        rep movsw
        int RMCALL_VECT
        mov esi,v86r_edi
        mov edi,[image]
        add edi,REGSTRUCOFF
        mov ecx,25
        rep movsw
        popfd
        pop es
        pop ds
        pop edi
        pop esi
        pop ecx
        iretd

;
modulegetmem:
        push edi
        push ds
        mov ds,[cs:data32sel]
        mov edi,[image]
        push eax
        mov eax,[edi+lomembaseOFF]
        add eax,edi
        mov [lomembase],eax
        mov eax,[edi+himembaseOFF]
        add eax,edi
        mov [himembase],eax
        pop eax
        call getmem
        pushfd                          ; "sub eax,edi" modifies CF
        push eax
        mov eax,[lomembase]
        sub eax,edi
        mov [edi+lomembaseOFF],eax
        mov eax,[himembase]
        sub eax,edi
        mov [edi+himembaseOFF],eax
        pop eax
        sub eax,edi
        popfd
        pop ds
        pop edi
        retf

;
modulegetlomem:
        push edi
        push ds
        mov ds,[cs:data32sel]
        mov edi,[image]
        push eax
        mov eax,[edi+lomembaseOFF]
        add eax,edi
        mov [lomembase],eax
        pop eax
        call getlomem
        pushfd
        push eax
        mov eax,[lomembase]
        sub eax,edi
        mov [edi+lomembaseOFF],eax
        pop eax
        sub eax,edi
        popfd
        pop ds
        pop edi
        retf

;
modulegethimem:
        push edi
        push ds
        mov ds,[cs:data32sel]
        mov edi,[image]
        push eax
        mov eax,[edi+himembaseOFF]
        add eax,edi
        mov [himembase],eax
        pop eax
        call gethimem
        pushfd
        push eax
        mov eax,[himembase]
        sub eax,edi
        mov [edi+himembaseOFF],eax
        pop eax
        sub eax,edi
        popfd
        pop ds
        pop edi
        retf

;
modulelomemsize:
        push edi
        push ds
        mov ds,[cs:data32sel]
        mov edi,[image]
        push eax
        mov eax,[edi+lomembaseOFF]
        add eax,edi
        mov [lomembase],eax
        mov eax,[edi+lomemtopOFF]
        add eax,edi
        mov [lomemtop],eax
        pop eax
        call lomemsize
        pop ds
        pop edi
        retf

;
modulehimemsize:
        push edi
        push ds
        mov ds,[cs:data32sel]
        mov edi,[image]
        push eax
        mov eax,[edi+himembaseOFF]
        add eax,edi
        mov [himembase],eax
        mov eax,[edi+himemtopOFF]
        add eax,edi
        mov [himemtop],eax
        pop eax
        call himemsize
        pop ds
        pop edi
        retf

;
modulegetdsc:
        push ds
        mov ds,[cs:data32sel]
        call getdsc
        pop ds
        retf

;
modulesetdsc:
        push ds
        mov ds,[cs:data32sel]
        call setdsc
        pop ds
        retf

;
modulegetvect:
        push ds
        mov ds,[cs:data32sel]
        call getvect
        pop ds
        retf

;
modulesetvect:
        push ds
        mov ds,[cs:data32sel]
        call setvect
        pop ds
        retf

;
modulemap_phys_addr:
        push edi
        push ds
        mov ds,[cs:data32sel]
        mov edi,[image]
        push eax
        mov eax,[edi+himembaseOFF]
        add eax,edi
        mov [himembase],eax
        pop eax
        call map_phys_addr
        pushfd
        push eax
        mov eax,[himembase]
        sub eax,edi
        mov [edi+himembaseOFF],eax
        pop eax
        popfd
        pop ds
        pop edi
        retf

;
moduledosshell:
        push eax
        push edi
        push ds
        mov ds,[cs:data32sel]
        mov edi,[image]
        mov eax,[edi+lomembaseOFF]
        add eax,edi
        mov [lomembase],eax
        cmp esi,0
        mov eax,0
        je .preok
        add esi,edi
        mov eax,-1
.preok:
        call dosshell
        pushfd
        cmp eax,0
        je .postok
        sub esi,edi
.postok:
        mov eax,[lomembase]
        sub eax,edi
        mov [edi+lomembaseOFF],eax
        popfd
        pop ds
        pop edi
        pop eax
        retf

;
moduledosprint:
        push ds
        mov ds,[cs:data32sel]
        add edx,[image]
        call dosprint
        sub edx,[image]
        pop ds
        retf

;
moduleputstr:
        push ds
        mov ds,[cs:data32sel]
        add esi,[image]
        call putstr
        sub esi,[image]
        pop ds
        retf

resb    16
