; This program is made by Daniel Horchner.
; email: Daniel.Horchner@95.Student.WAU.NL
;        dbjh@usa.net

segment code32 use32

%include "raw32.inc"

global  main

;32-bit data
TSS1:
                TSS
                times 2000h db 0        ; 1 bit for each port; 64K / 8 = 8K
TSS2:
                TSS
                times 2000h db 0

LDTdsc:
                seg_descriptor  0,0,0, 82h, 0           ; dsc type 2=LDT
LDTsel          dw      0
TSS1dsc:
                seg_descriptor  2067h,0,0, 89h, 0       ; dsc type 9=TSS
TSS2dsc:
                seg_descriptor  2067h,0,0, 89h, 0       ; 68h bytes for task
TSS1sel         dw      0                               ;  state + 2000h
TSS2sel         dw      0                               ;  bytes for I/O map

%macro  gate_dsc 0-5 0,0,0,0,0
  .offset0_15   dw      %1
  .selector     dw      %2
  .count        db      %3
  .access       db      %4
  .offset16_31  dw      %5
%endmacro

callgatedsc:
                gate_dsc        0,0, 0, 8ch, 0          ; dsc type ch=call
returnsel       dw      0                               ;  gate
org_tr          dw      0               ; original Task Register contents
gdtr            dw      0,0,0           ; 16-bit limit, 32-bit base
LDT             dd      0               ; pointer to mem allocated for LDT

msg0            db      'Not running at Privilege Level 0.','$'
msg1            db      'Press a key to continue.',0
msg2            db      'Message from procedure!',0
msg3            db      'Message from task2!',0

;32-bit code
main:
                                        ; First, check if running at PL 0
        mov ebx,cs
        lar ecx,ebx
        and ecx,6000h                   ; Only the DPL bits are needed
        jz short .PL0
        mov edx,msg0
        call dosprint
        jmp @exit
.PL0:
;
                                        ; Make a copy of the GDT to create a
        sgdt [gdtr]                     ;  LDT; Make all descriptors PL 3
        movzx ecx,word [gdtr]
        mov [LDTdsc.limit0_15],cx       ; ecx=GDT limit in bytes
        inc ecx                         ; ecx=GDT size in bytes
        mov eax,ecx
        call getlomem                   ; LOmem -> Handy when using TD ;)
        mov [LDT],eax                   ; Save ptr to mem allocated for LDT
        mov edi,eax
        push ds
        pop es               
        mov esi,[gdtr+2]
        mov ds,[zerosel]
        rep movsb
        mov ds,[cs:data32sel]           ; Restore ds
                                        ; Change the DPL field of the
        movzx ecx,word [gdtr]           ;  descriptors to PL 3
        inc ecx
        shr ecx,3                       ; ecx=number of selectors in GDT/LDT
        mov edi,[LDT]
next_descriptor:
        mov ebx,ecx
        dec ebx
        or byte [edi+ebx*8+5],3 << 5
        loop next_descriptor            ; PL 3; DPL bits are bits 5 and 6

        mov cx,1                        ; Allocate 1 descriptor for LDT
        call getdsc
        jc near @exit
        mov [LDTsel],ax

        mov eax,[code32a]
        add eax,[LDT]
        mov [LDTdsc.base0_15],ax
        shr eax,16
        mov [LDTdsc.base16_23],al
        mov [LDTdsc.base24_31],ah

        mov edi,LDTdsc                  ; es:edi=pointer to 8 byte dsc buffer
        mov bx,[LDTsel]                 ; bx=selector
        call setdsc
        jc near @exit

        lldt [LDTsel]
;
; Switch to a task running at a Privilege Level > 0
        mov cx,2                        ; Allocate 2 descriptors for TSS's
        call getdsc
        jc near @exit
        mov [TSS1sel],ax
        add ax,8
        mov [TSS2sel],ax

        mov eax,[code32a]
        add eax,TSS1
        mov [TSS1dsc.base0_15],ax
        shr eax,16
        mov [TSS1dsc.base16_23],al
        mov [TSS1dsc.base24_31],ah
        mov eax,[code32a]
        add eax,TSS2
        mov [TSS2dsc.base0_15],ax
        shr eax,16
        mov [TSS2dsc.base16_23],al
        mov [TSS2dsc.base24_31],ah

        mov edi,TSS1dsc
        mov bx,[TSS1sel]                ; bx=selector
        call setdsc
        jc near @exit
        mov edi,TSS2dsc
        mov bx,[TSS2sel]
        call setdsc
        jc near @exit

        str [org_tr]                    ; Save original Task Register
        cli

        mov eax,cr3                     ; CR3 and LDTR are loaded but not
        mov [TSS1.@cr3],eax             ;  stored on a task switch
        mov [TSS2.@cr3],eax
        sldt [TSS1.ldtr]
        sldt [TSS2.ldtr]
        ltr [TSS1sel]                   ; TSS1sel = current task

        pushfd
        pop eax
        mov [TSS2.eflags],eax
        mov eax,ds
        or eax,7                        ; Set TI bit; Set RPL to 3
        mov [TSS2.@ds],eax
        mov [TSS2.@es],eax
        mov [TSS2.@fs],eax
        mov [TSS2.@gs],eax
        mov [TSS2.@ss],eax
        mov [TSS2.ss0],ss               ; Save PL 0 ss
        mov [TSS2.esp0],esp             ; Save PL 0 esp
        mov eax,200h                    ; Allocate a 200h stack
        call gethimem
        jc near @exit
        add eax,200h                    ; Stack grows down...
        mov [TSS2.@esp],eax
        mov eax,cs
        or eax,7                        ; Set TI bit; Set RPL to 3
        mov [TSS2.@cs],eax
        mov dword [TSS2.eip],task2
        call far [TSS2sel-4]            ; Switch to other task ('task2:')
;
; Call a procedure at a Privilege Level > 0 and return to PL 0 using a call
;  gate
        mov cx,1                        ; Allocate 1 descriptor for call gate
        call getdsc
        jc near @exit
        mov [returnsel],ax

        mov [callgatedsc.selector],cs
        mov word [callgatedsc.offset0_15],return_to_PL0
        or byte [callgatedsc.access],3 << 5 ; PL 3 and higher PL code must be
                                        ;  able to call it
        mov edi,callgatedsc
        mov bx,[returnsel]              ; bx=selector
        call setdsc
        jc near @exit

        push ds
        push es
        push fs
        push gs
        push ss
                                        ; The PL 0 ss:esp will be fetched
        mov [TSS1.ss0],ss               ;  from the current TSS after the
        mov [TSS1.esp0],esp             ;  interlevel CALL

        mov eax,ss                      ; Since cs on stack will be a sel in
        or eax,7                        ;  the LDT with PL > 0 IRETD will be
        push eax                        ;  an interlevel one -> ss:esp are
        push esp                        ;  also popped from the stack; ss's
        pushfd                          ;  RPL must be equal to cs's RPL
        mov eax,cs
        or eax,7                        ; Set TI bit; Set RPL to 3
        push eax
        push dword procedure
                                        ; Change all segment registers for
        mov eax,ds                      ;  their PL 3 equivalent in the LDT
        or eax,7                        ;  If they are not made PL 3 all
        mov ds,eax                      ;  these regs will be set to zero
        mov eax,es
        or eax,7                        ; Set Table Indicator bit; Set RPL to
        mov es,eax                      ;  3
        mov eax,fs
        or eax,7
        mov fs,eax
        mov eax,gs
        or eax,7
        mov gs,eax

        iretd                           ; Jump to 'procedure:'
return_to_PL0:
        add esp,16                      ; Remove ss:esp and cs:eip from stack
        pop ss                          ;  (result of call to call gate)
        pop gs
        pop fs
        pop es
        pop ds
;
        mov eax,[gdtr+2]
        movzx ebx,word [org_tr]
        and ebx,~ 3                     ; ebx=offset of TSS dsc in GDT
        and byte [gs:eax+ebx+5],~ 2     ; Clear Busy bit
        ltr [org_tr]                    ; Restore original Task Register
        sti

        mov esi,msg1
        @rlp edi,0b8000h+3*160
        mov bl,1fh
        call putstr
        mov byte [v86r_ah],0            ; ah=0 -> Wait for key and read char
        mov al,16h
        int RMCALL_VECT

@exit:
        jmp exit                        ; Return to real/V86 mode

;
procedure:
        mov esi,msg2
        @rlp edi,0b8000h+160
        mov bl,1fh
        call putstr
;        mov eax,cr0                     ; Exception 13 if this is PL > 0 code
        call far [returnsel-4]          ; This is allowed because the offset
                                        ;  is discarded
;
task2:
        mov esi,msg3
        @rlp edi,0b8000h+2*160
        mov bl,1fh
        call putstr
;        mov eax,cr0                     ; Exception 13 if this is PL > 0 code
        iretd
;
get_eip:
        mov eax,[esp]
        ret

resb    16
