; This program is made by Daniel Horchner. Based on Thomas Pytel's START32 /
; Adam Seychell's DOS32 v1.2 (Who was first?)
; Read RAW32.TXT!
; email: Daniel.Horchner@95.Student.WAU.NL
;        dbjh@usa.net

        .386p
        locals

;RMCB_ON         =       0       ; If defined -> IRQs in RM reflected to PM
NO_EXTERN       =       0       ; 'extrn' directives must not be included
include raw32.inc
extrn   main:near
public  DPMImem,VCPImem,XMSmem  ; Made public for debugging purposes

LOWMIN          =       5       ; minimum conventional memory needed (in KB)
HIGHMIN         =       12      ; minimum extended memory needed (in KB)
RM_STACKSIZE    =       200h    ; real mode (interrupt) stack size (in bytes)
PM_STACKSIZE    =       200h    ; protected mode stack size (in bytes)
WINFRIENDLY     =       0       ; If defined -> Allocate only half of the
                                ;  possible DPMI mem under Windows
DPMImeminfostruc        struc
  largest_free          dd      0
  max_unlocked          dd      0
  max_locked            dd      0
  lin_addr_space        dd      0
  tot_unlocked          dd      0
  tot_free              dd      0
  tot_physical          dd      0
  free_lin_addr_space   dd      0
  pagingfilesize        dd      0
  reserved              db      12 dup(0ffh)
DPMImeminfostruc        ends

checkifexc      macro   intn
        push eax
        mov al,0bh                      ; Operation Control Word 3: bit 0=1
        out 20h,al                      ;  -> IRR on next read; bit 1=1 ->
        in al,20h                       ;  Act on value of bit 0
        test al,(1 shl (intn-8))        ; Interrupt Request Register contains
        pop eax                         ;  bit vector which IRQs await service
        jnz short @@isIRQ
        jmp fword ptr cs:exc&intn&handler ; No IRQ -> Jump to exc. handler
@@isIRQ:
endm

code16  segment para public use16
        assume cs:code16, ds:code16, ss:code16

;16-bit data
org RM_STACKSIZE                        ; real mode stack
idtrRM          dw      3ffh,0,0
stackptrRM      label   dword
spRM            dw      RM_STACKSIZE/2  ; Use lower half of RM stack for IRQs
ssRM            dw      code16

errmsg0         db      'General error.$'
errmsg1         db      'Not enough conventional memory (free).$'
errmsg2         db      'Not enough extended memory (free).$'
errmsg3         db      'Could not enable A20 gate.$'
errmsg4         db      'Could not disable A20 gate.$'
errmsg5         db      'DPMI error: $'
errmsg6         db      'VCPI error: $'
errmsg7         db      'XMS error: $'
errmsg8         db      'No way found to enter protected mode.$'
errmsg9         db      'Conflict with PIC mapping.$'

DPMIcall        dd      0
DPMImem_hst     dw      0               ; mem needed by DPMI host (in paras)
DPMImem_hst_seg dw      0               ; segment of mem to provide to host
DPMImeminfo     DPMImeminfostruc        <>
DPMIhandle      dd      0               ; handle of allocated DPMI mem block
VCPIpage_base   dd      0
VCPIpage_top    dd      0
XMScall         dd      0
XMShandle       dw      0               ; handle of allocated XMS mem block
SXMSbit         db      0
gate            db      0
enterPM         dw      raw_PM
                dw      raw_PM
                dw      VCPI_PM
                dw      DPMI_PM
orgIVT          db      16*4 dup(0)     ; vectors to orginal RM IRQs

;16-bit code
check_memory:
                                        ; Get conventional memory size
        movzx eax,word ptr ds:pspbuffer[2] ; ax=top of DOS memory in paras
        shl eax,4                       ; eax=amount of conventional memory
        sub eax,ds:code32a    ; eax=mem below 640KB above start code32
        mov ds:lomemtop,eax   ;   =offset last byte low mem + 1 from code32
        sub eax,ds:lomembase  ; eax=mem below 640KB above end code32
        shr eax,10            ; eax=the same, but in KB (=free conv mem)
        cmp ax,LOWMIN
        jb notenoughlowmem
                                        ; Get extended memory size
        clc                             ; Ensure Carry Flag set only on error
        mov ax,code32end
        mov es,ax
        xor di,di                       ; es:di=pointer to buffer for result
        mov edx,'SMAP'
        xor ebx,ebx                     ; ebx=0 -> Start at begin of map
@@fillbuffer:
        mov ecx,20                      ; ecx=num of bytes to copy to buffer
        mov eax,0e820h                  ; eax=e820h -> Get system memory map
        int 15h
        jc short @@fn_e801
        add di,cx                       ; cx=number of bytes copied
        cmp ebx,0                       ; ebx=next offset to copy from or 0
        jnz short @@fillbuffer          ;  if ready

        mov cx,di
        xor di,di                       ; di -> 1st Phoenix BIOS system mem
        xor eax,eax                     ;  map address range descriptor
@@next_descriptor:
        cmp dword ptr es:[di+16],1      ; descriptor+16=type of address range
        jne short @@loop                ;  1 -> Mem available to OS
        mov ebx,es:[di+8]               ; descriptor+8 (qword)=length in bytes
        cmp es:[di],100000h             ; descriptor+0 (qword)=base address
        jb short @@loop                 ;  extended mem = mem above 1MB ;)
        add eax,ebx
        cmp ds:himembase,0
        jne short @@loop
        mov ebx,es:[di]                 ; descriptor+0 (qword)=base address       
        mov ds:himembase,ebx
@@loop:
        add di,20                       ; 20 = descriptor size in bytes
        cmp di,cx
        jb short @@next_descriptor

        shr eax,10                      ; eax=total extended memory in KB
        jmp short @@exit

@@fn_e801:
        clc                             ; Clear CF of previous failed call
        mov ax,0e801h                   ; ax=e801h -> Get memory size
        int 15h
        jc short @@fn_c7
        and edx,0ffffh                  ; edx=configured mem > 16M in 64KB
        shl edx,6                       ; edx=configured mem > 16M in KB
        movzx eax,cx                    ; eax=configured mem 1-16M in KB
        add eax,edx                     ; eax=total extended memory in KB
        jmp short @@exit
@@fn_c7:
        clc                             ; Clear CF of previous failed call
        mov ah,0c0h                     ; ah=c0h -> Get configuration
        int 15h                         ; es:bx=pointer to ROM table
        jc short @@fn_88readCMOS
        test byte ptr es:[bx+6],10h     ; If bit 4 in feature byte 2 is set ->
        jz short @@fn_88readCMOS        ;  int 15h/ah=c7h is supported
        push ds
        mov ax,code32end
        mov ds,ax
        xor si,si                       ; ds:si=pointer to buffer
        mov ah,0c7h                     ; ah=c7h -> Return memory map info
        int 15h
        mov eax,ds:[0ah]
        add eax,ds:[0eh]                ; eax=total extended memory in KB
        pop ds
        jmp short @@exit

; If int 15h/ah=88h is hooked it may not return the total amount of ext mem
; Function 88h reads from CMOS addresses 30h and 31h (=1 word -> max 64MB)
@@fn_88readCMOS:
        mov al,31h              ; Write CMOS addr to read/write to port 70h
        out 70h,al              ; Read/write port 71h to get/set data
        in al,71h                       ; 31h -> MSB of total ext mem in KB
        mov ah,al       
        mov al,30h                      ; 30h -> LSB of total ext mem in KB
        out 70h,al 
        in al,71h       
        and eax,0ffffh                  ; eax=total extended memory in KB
        jnz short @@exit
@@fn_88:
;       clc                             ; CF is cleared by "and eax,0ffffh"
        mov ah,88h                      ; ah=88h -> Get extended memory size
        int 15h
        jc short @@errexit
        and eax,0ffffh                  ; eax=total extended memory in KB
@@exit:
        cmp eax,HIGHMIN
        jb short notenoughhighmem
        mov ds:totalextmem,eax
        ret
@@errexit:
        mov dx,offset errmsg0           ; The total amount of extended memory
        mov ds:returncode,0ffh          ;  could not be determined
        jmp exit16err

notenoughlowmem:
        mov dx,offset errmsg1
        mov ds:returncode,0ffh
        jmp exit16err
notenoughhighmem:
        mov dx,offset errmsg2
        mov ds:returncode,0ffh
        jmp exit16err
notenoughhighmemDPMI:                   ; Entry point if not enough DPMI mem
        mov ds:v86r_dx,offset errmsg2   ;  free; this code runs in 16-bit PM
        mov ds:returncode,0ffh
        jmp exit16errDPMI
;
buildIDT:
        xor eax,eax                     ; Ensure high word of eax is zero
        mov dx,offset errmsg9           ; First, check if there is no
        mov ds:returncode,0ffh          ;  conflict between PIC bases and the
        mov al,ds:mPICbase              ;  place of RMCALL_VECT in the IDT
        cmp al,RMCALL_VECT+N_INT        ;  and if there are N_INT gates free
        ja short @@mPICbaseok           ;  above RMCALL_VECT
        cmp al,RMCALL_VECT-7
        jnb exit16err
@@mPICbaseok:
        mov bl,ds:sPICbase
        cmp bl,RMCALL_VECT+N_INT
        ja short @@PICbasesok
        cmp bl,RMCALL_VECT-7
        jnb exit16err
@@PICbasesok:

        cmp al,bl
        ja short @@basedone             ; Highest base determines IDT size
        mov al,bl
@@basedone:
        add ax,7                        ; ax=hightest int number for any IRQ
        cmp ax,RMCALL_VECT+N_INT        ; IDT large enough for RMCALL_VECT?
        ja short @@IDTsizeok
        mov ax,RMCALL_VECT+N_INT        ; Adjust IDT size
@@IDTsizeok:
        inc ax                          ; ax=number of needed IDT entries
        mov dx,ax                       ; Save number of needed IDT entries 

        mov ebx,ds:lomembase            ; Align lomembase on paragraph
        add ebx,ds:code32a
        add ebx,0fh
        and bl,0f0h
        mov dword ptr ds:idtr[2],ebx    ; Set IDT base
        mov ecx,ebx
        shr ebx,4
        mov es,bx                       ; es=segment of IDT
        sub ecx,ds:code32a
        mov ds:idt,ecx                  ; idt=offset IDT from code32

        shl ax,3                        ; ax=ax*8 = amount mem needed for IDT
        add ds:lomembase,eax
        dec ax
        mov ds:idtr[0],ax               ; Set IDT limit (=size-1)

        mov ax,code32dsc-gdt            ; ax=code32sel
        shl eax,16
        mov ax,small offset unexp       ; eax=LSD of int gate to unexp int
        xor di,di
        mov cx,dx
@@nextLSD:                              ; Set least significant dword of all
        mov es:[di],eax                 ;  IDT entries
        add di,8
        loop @@nextLSD
        mov eax,8e00h                   ; eax=MSD of int gate (Present, DPL 0)
        mov di,4
        mov cx,dx
@@nextMSD:                              ; Set most significant dword of all
        mov es:[di],eax                 ;  IDT entries
        add di,8
        loop @@nextMSD

        mov ax,code32
        mov fs,ax                       ; offset table resides in code32
        mov cx,8                        ; Set exc8-15/IRQ0-7 handler offsets
        movzx di,ds:mPICbase
        cmp di,8                        ; int shared between excs and IRQs?
        jne short notshared
        shl di,3                        ; di=offset of IDT entry
        mov bx,small offset intoff
@@next_entry:
        mov ax,fs:[bx]
        mov es:[di],ax
        add bx,2
        add di,8
        loop @@next_entry
        mov ax,small offset exc16       ; Set exc16 handler offset
        mov es:[di],ax
        jmp set_remaining

notshared:                              ; First, set IRQ0-7 handler offsets
        shl di,3
        mov bx,small offset IRQoff
@@next_IRQ:
        mov ax,fs:[bx]
        mov es:[di],ax
        add bx,2
        add di,8
        loop @@next_IRQ

        mov cx,9                        ; Set exc8-16 handler offsets
        mov di,8*8
        mov bx,small offset excoff + 8*2
@@next_exc:
        mov ax,fs:[bx]
        mov es:[di],ax
        add bx,2
        add di,8
        loop @@next_exc

set_remaining:                          ; IRQ8-15 and exc0-7 handler offsets
        mov bx,small offset IRQoff + 8*2
        mov cx,8
        movzx di,ds:sPICbase
        shl di,3
@@next_IRQ:
        mov ax,fs:[bx]
        mov es:[di],ax
        add bx,2
        add di,8
        loop @@next_IRQ

        mov bx,small offset excoff      ; Set exc0-7 handler offsets
        mov cx,8
        xor di,di
@@next_exc:
        mov ax,fs:[bx]
        mov es:[di],ax
        add bx,2
        add di,8
        loop @@next_exc

        ret
;
; This routine is always needed under non-DPMI, because it initializes orgIVT
init_callbacks:
        mov ax,0
        mov es,ax                       ; es=segment of RM IVT
        mov ax,code32
        mov fs,ax                       ; offset table resides in code32
        mov si,offset orgIVT
        mov ax,code16
        shl eax,16
        mov di,8*4                      ; offset in IVT of vector for int 8
        mov bx,small offset callbackoff
        mov cx,8
@@next_callback1:
        mov edx,es:[di]                 ; Read vector of orginal IRQ handler
        mov [si],edx                    ; Save vector
ifdef   RMCB_ON
        cmp di,15*4                     ; Problems with int 15 (not IRQ 7!?)
        je short @@callback1done        ;  -> Don't install a RMCB for it
        mov ax,fs:[bx]                  ; Read offset of new IRQ handler
        mov es:[di],eax                 ; Install new IRQ handler (callback)
@@callback1done:
endif
        add di,4
        add si,4
        add bx,2
        loop @@next_callback1

        mov di,70h*4                    ; offset in IVT of vector for int 70h
        mov cx,8
@@next_callback2:
        mov edx,es:[di]
        mov [si],edx
ifdef   RMCB_ON
        cmp di,76h*4                    ; Problems with IRQ 14 -> Don't
        je short @@callback2done        ;  install a real mode callback for it
        mov ax,fs:[bx]
        mov es:[di],eax
@@callback2done:
endif
        add di,4
        add si,4
        add bx,2
        loop @@next_callback2
        ret
;
restoreIVT:
ifdef   RMCB_ON
        mov ax,0
        mov es,ax
        mov si,offset orgIVT
        mov di,8*4
        mov cx,8
@@next_vector1:
        lodsd
        stosd
        loop @@next_vector1

        mov di,70h*4
        mov cx,8
@@next_vector2:
        lodsd
        stosd
        loop @@next_vector2
endif
        ret
;
check_V86:                              ; Check if running in V86 mode
        smsw ax                         ; SMSW is not PL sensitive (CR0 is)
        test al,1                       ; If Protection Enable bit is set ->
        jz short @@not_V86              ;  running in protected mode; prog
@@V86:                                  ;  runs under DOS -> must be V86 mode
        stc
        ret
@@not_V86:
        clc
        ret
;
check_windows:
        mov ax,4680h                    ; ax=4680h -> Windows 3.0 installation
        int 2fh                         ;  check
        cmp ax,0                        ; ax=0 -> Windows 3.0 running
        je short @@windows
        mov ax,1600h                    ; al=0 -> Get installed state, ah=16h
        int 2fh                         ;  -> Windows 386
        test al,7fh                     ; al=1/ffh/version number -> Windows
        jnz short @@windows             ; al=80h -> XMS 1
@@no_windows:
        mov ds:windows,0
        ret
@@windows:
        mov ds:windows,1
        ret
;
rawinit:
        call buildIDT                   ; es=segment of IDT
        mov word ptr es:[RMCALL_VECT*8],small offset RMcall_int

        call enableA20

        cmp ds:himembase,0              ; Check if 'check_memory' has already
        jne short @@himembase_done      ;  set himembase
        mov ds:himembase,100000h        ; =address of begin mem block
@@himembase_done:
        mov eax,ds:code32a
        sub ds:himembase,eax            ; =offset start block from code32

        mov eax,ds:totalextmem
        shl eax,10                      ; eax=amount of extended mem in bytes
        add eax,ds:himembase
        mov ds:himemtop,eax             ; =offset of highest byte + 1 in mem
                                        ;  from start code32
        call init_callbacks
        
        mov ax,ds:code16sel
        mov word ptr ds:exitPM32[4],ax
        mov ds:exitPM32[0],offset rawexit

        ret
;
enableA20:
        mov gate,2                      ; 2 -> Enable A20
        call gateA20
        jc short @@notenabledA20
        ret
@@notenabledA20:
        mov dx,offset errmsg3
        mov ds:returncode,0ffh
        jmp exit16err
;
disableA20:
        mov gate,0                      ; 0 -> Disable A20
        call gateA20
        jc short @@notdisabledA20
        ret
@@notdisabledA20:
        mov dx,offset errmsg4
        mov ds:returncode,0ffh
        jmp exit16err
;
gateA20:
        xor ax,ax                       ; Set fs to 0 and gs to ffffh for
        mov fs,ax                       ;  testA20
        dec ax
        mov gs,ax

        call testA20                    ; Is A20 enabled or disabled?
        call isA20ok                    ; Is that right?
        jnc @@gateA20done               ; If yes, done

        in al,92h                       ; PS/2 A20 gate
        and al,0fdh                     ; Clear bit 1 (=gate A20)
        or al,gate                      ; Set A20 bit or leave it 0
        out 92h,al                      ; 2 -> Enable A20; 0 -> Disable A20

        call testA20                    ; Is A20 enabled or disabled?
        call isA20ok                    ; Is that right?
        jnc short @@gateA20done         ; If yes, done

                                        ; A20 gate via 8042 (=keyb controller)
        call empty_8042                 ; Ensure 8042 input buf is empty prior
        jnz short @@gateA20fail         ;  to writing to the input registers
        mov al,0adh                     ; adh -> Disable keyboard interface
        out 64h,al                      ; 64h = 8042 command register

        call empty_8042
        jnz short @@gateA20fail
        mov al,0d0h                     ; d0h -> Read 8042 output port; value
        out 64h,al                      ;  in output port is placed in output
                                        ;  register
        xor cx,cx
@@wait8042:
        in al,64h                       ; Read 8042 status register
        test al,1                       ; Output reg (60h) has data for sys?
        loopz @@wait8042                ; If not (yet) -> Loop
        jz short @@gateA20fail

        in al,60h                       ; Data in 8042 output reg -> Read it
        push ax                         ; Save contents of 8042 output port

        call empty_8042
        pop ax                          ; If error -> valid return address
        jnz short @@gateA20fail         ;  on stack
        push ax
        mov al,0d1h                     ; d1h -> Write to output port of 8042
        out 64h,al                      ;  (through port 60h)
                                  
        call empty_8042
        pop ax
        jnz short @@gateA20fail
        and al,0fdh                     ; Clear bit 1 (=gate A20)
        or al,gate                      ; Set A20 bit or leave it 0
        out 60h,al                      ; 60h = 8042 output port (now)

        call empty_8042
        jnz short @@gateA20fail
        mov al,0aeh                     ; aeh -> Enable keyboard interface
        out 64h,al

                                        ; Wait for 8042 to gate A20
        mov al,36h                      ; Select counter 0; write LSB, MSB;
        out 43h,al                      ;  cntr mode 3; 16-bit binary counter
        mov al,0                        ; Set counter value to 65536; no check
        out 40h,al                      ;  is necessary if wrap takes place ->
        out 40h,al                      ;  16-bit register wrap does the job
                                        ; 65536 is default counter value too
        mov cx,30                       ; Wait maximum 30 ticks:
@@anothertick:                          ;  30 * 1/(1193181 ticks/s)  25 s
        mov al,0d2h                     ; Read back command; bit 5=0 -> Latch
        out 43h,al                      ;  counter; bit 4!=0 -> Don't latch
        in al,40h                       ;  status; bit 1=1 -> Select cntr 0;
        mov dl,al                       ;  bit 0 reserved. Writing 0 instead
        in al,40h                       ;  of read back command also works.
        mov dh,al
@@nexttick:                             ; Wait at least 1 PIT clock tick
        mov al,0d2h
        out 43h,al
        in al,40h                       ; Get LSB of current tick counter
        mov ah,al
        in al,40h                       ; Get MSB of current tick counter
        xchg ah,al                      ; If ax is greater than dx (instead
        cmp ax,dx                       ;  of the opposite), dx will still
        je short @@nexttick             ;  contain the right value because of
        sub dx,ax                       ;  16-bit reg wrap (e.g. 1-ffff=2)
        sub cx,dx                       ; Substract the number of passed
        jnc short @@anothertick         ;  clock ticks between 2 port reads

        call testA20                    ; Is A20 enabled or disabled?
        call isA20ok                    ; Is that right?
        jnc short @@gateA20done

@@gateA20fail:
        stc                             ; Error -> Carry Flag set
@@gateA20done:
        ret
;
testA20:
        mov al,fs:[0]                   ; Get byte from 0:0
        mov ah,al                       ; Save byte
        not al                          ; Modify byte
        xchg al,gs:[10h]                ; If A20 is off -> ffffh:10h = 0:0
        cmp ah,fs:[0]                   ; 0:0 has changed only if A20 is off
        mov gs:[10h],al                 ; Restore byte at ffffh:10h
        ret                             ; Zero Flag is set if A20 is enabled
;
isA20ok:                                ; Carry Flag set = A20 is not ok
        pushf                           ; Save result, zero -> A20 is on
        test gate,2                     ; Must A20 be enabled or disabled?
        jz short @@disableA20           ; If gate != 2 -> must be disabled
        popf
        clc                             ; Ensure Carry Flag set only on error
        jz short @@ok                   ; Must be on, it is on -> A20 is ok
        jmp short @@notok
@@disableA20:
        popf
        clc
        jnz short @@ok                  ; Must be off, it is off -> A20 is ok
        jmp short @@notok
@@notok:
        stc                             ; Set Carry Flag
@@ok:
        ret                             ; Leave Carry Flag cleared
;
empty_8042:                             ; Wait for safe to write to 8042
        xor cx,cx                       ;  (= keyboard controller)
@@try_8042:
        in al,64h                       ; Read 8042 status register
        test al,2                       ; Input reg (60/64) has data for 8042?
        loopnz @@try_8042               ; If yes -> Loop
        ret
;
check_XMS:
        mov ax,4300h                    ; ax=4300h -> Check XMS installation
        int 2fh
        cmp al,80h                      ; XMS available?
        je short @@XMS
@@no_XMS:
        stc
        ret
@@XMS:
        clc
        ret
;
XMSinit:
        cmp ds:systemtype,2             ; Check if called by VCPIinit. If so
        je short @@VCPIinit             ;  -> already build IDT
        call buildIDT                   ; es=segment of IDT
        mov word ptr es:[RMCALL_VECT*8],small offset RMcall_int
@@VCPIinit:

        mov ax,4310h                    ; ax=4310h -> Get XMS driver entry
        int 2fh                         ;  point
        mov word ptr XMScall[0],bx      ; bx=entry point offset
        mov word ptr XMScall[2],es      ; es=entry point segment

        mov ah,0                        ; ah=0 -> Get XMS version number
        call XMScall
        cmp ah,3                        ; ah=XMS version number (major)
        jb short @@enableA20            ; SXMS (4GB) available from XMS 3
        mov SXMSbit,80h                 ; Call with bit 7 set in al -> SXMS

@@enableA20:
        mov ah,5                        ; ah=5 -> Local enable A20 line
        call XMScall
        cmp al,0                        ; Error?
        mov dx,offset errmsg3
        mov ds:returncode,0ffh
        je exit16err

        mov eax,800h                    ; ah=8 -> Query free extended memory
        or ah,SXMSbit                   ;  (Clear high word for if no SXMS)
        call XMScall                    ; When high bit of bl is set ->
        test bl,80h                     ;  bl=error code
        mov dx,offset errmsg7           ; eax=largest block size available in
        mov ds:returncode,1             ;  KB
        jnz exit16err                   ; edx=total extended memory in KB
        mov ds:XMSmem,eax
        cmp eax,HIGHMIN
        jb notenoughhighmem
;       cmp eax,0                       ; Not necessary; XMS driver returns
;       je notenoughhighmem             ;  an error when no free XMS mem left
        mov ds:himemtop,eax             ; =now only _size_ of mem block
        shl ds:himemtop,10

        mov edx,eax                     ; edx=amount of ext mem wanted, in KB
        mov ah,9                        ; ah=9 -> Allocate extended mem block
        or ah,SXMSbit
        call XMScall                    ; dx=block handle
        mov XMShandle,dx
        cmp al,0                        ; Error?
        mov dx,offset errmsg7
        mov ds:returncode,2
        je exit16err

        mov dx,XMShandle                ; dx=block handle
        mov ah,0ch                      ; ah=0ch -> Lock extended mem block
        call XMScall
        push dx                         ; Save dx for when no error
        cmp al,0                        ; Error?
        mov dx,offset errmsg7
        mov ds:returncode,3
        je exit16err
                                        ; dx:bx=32-bit physical block address
        pop dx                          ;  after dx has been restored
        shrd eax,edx,16                 ; Physical block address, high word
        mov ax,bx                       ; Physical block address, low word
        sub eax,ds:code32a
        mov ds:himembase,eax            ; =offset start block from code32
        add ds:himemtop,eax             ; =offset of 1st byte above block
                                        ;  from start code32

        cmp ds:systemtype,2
        je short @@VCPIinit2
        call init_callbacks
@@VCPIinit2:

        mov ax,ds:code16sel
        mov word ptr ds:exitPM32[4],ax
        mov ds:exitPM32[0],offset XMSexit

        ret
;
check_VCPI:
        xor ax,ax
        mov es,ax
        cmp dword ptr es:[4*67h],0      ; Check if int 67h handler exists
        je short @@no_VCPI
        mov ax,0de00h                   ; al=0 -> VCPI installation check
        int 67h
        cmp ah,0                        ; No error?
        je short @@VCPI
@@no_VCPI:
        stc
        ret
@@VCPI:
        clc
        ret
;
VCPIinit:
        mov eax,0de0ah                  ; al=0ah -> Get 8259A int mappings
        int 67h
        mov ds:mPICbase,bl
        mov ds:sPICbase,cl

        mov ax,code32
        mov es,ax
        mov al,cl
        mov cx,8
        mov di,small offset IRQ8_vect
@@next_slaveIRQ:
        stosb
        inc al
        loop @@next_slaveIRQ
        mov cx,8
        mov al,bl
        mov di,small offset IRQ0_vect
@@next_masterIRQ:
        stosb
        inc al
        loop @@next_masterIRQ

        call buildIDT                   ; es=segment of IDT
        mov word ptr es:[RMCALL_VECT*8],small offset VCPIcall_int

        mov eax,ds:code32a
        add ds:VCPIidtr,eax
        add ds:VCPIgdtr,eax
                                        ; Set up page directory & page tables
        add eax,ds:lomembase
        add eax,0fffh                   ; Align page directory on 4K boundary
        and eax,0fffff000h
        mov ds:VCPIdir_base,eax
        shr eax,4
        mov es,ax

        mov ax,0de03h                   ; al=3 -> Get number of free 4KB pages
        int 67h
        push dx                         ; Save dx for when no error
        cmp ah,0                        ; No error?
        mov dx,offset errmsg6
        mov ds:returncode,1
        jne exit16err
        pop dx                          ; Restore dx (edx=number of pages)
        shl edx,2                       ; edx=free VCPI mem in KB
        mov ds:VCPImem,edx
        cmp edx,HIGHMIN
        jb notenoughhighmem
        cmp edx,0
        je notenoughhighmem             ; Check for 0KB free (due alloc algo)

                                        ; Clear page directory
        xor di,di
        mov cx,4096/4
        xor eax,eax
        rep stosd
                                        ; Get current 1st page table (0-4MB)
        mov di,4096                     ; es:di=ptr to 4KB page table buffer
        mov esi,offset ds:VCPIdsc0      ; ds:si is filled in by server (3 dsc)
        mov ax,0de01h                   ; al=1 -> Get PM interface
        int 67h
        cmp ah,0                        ; No error?
        mov dx,offset errmsg6           ; di=first free entry in page table
        mov ds:returncode,2
        jne exit16err
        mov ds:VCPIcall,ebx             ; ebx=offset server's PM entry point
                                        ;  (in VCPIdsc0)
        and edi,0ffffh                  ; High word is undefined -> Clear it
        mov VCPIpage_base,edi           ; Save entry address where own made
        mov eax,edi                     ;  page table entries start
        sub eax,4096                    ; eax=first free entry in page table
        shl eax,10                      ;    =number of used 1KB blocks
        sub eax,ds:code32a
        mov ds:himembase,eax            ; =offset start block from code32
                                        ;  (linear address)

                                        ; Allocate new pages
        call check_XMS                  ; Check for XMS to do XMS mem alloc
        jnc short @@XMSalloc            ;  for XMS mem allocation is _much_
                                        ;  faster than VCPI mem allocation.
                                        ;  XMS can be > 32768, many VCPI's not
                                        ;  -> Avail. mem not limited to VCPI

        mov cx,di                       ; cx=first free entry in page table
        add cx,3*4                      ; 3 pages; 1 for page dir, 1 for
@@next_page:                            ;  server's page, 1 to support 4GB
        mov ax,0de04h                   ; al=4 -> Allocate a 4KB page
        int 67h
        cmp ah,0                        ; No error?
        jne short @@last_page           ; edx=physical page address
        and edx,0fffff000h              ; Clear page attributes
        or edx,7                        ; User (=page addressable regardless
        mov es:[di],edx                 ;  the CPL); Read and write; Present
        add di,4
        cmp di,cx
        jb short @@next_page
@@last_page:                            ; edi=first free entry in page table
        cmp di,cx                       ; Check if the 3 pages are actually
        mov dx,offset errmsg6           ;  allocated or a page fault will be
        mov ds:returncode,3             ;  generated when in VCPIinitPM
        jb exit16err
        mov VCPIpage_top,edi
        jmp short @@setup_pagedir

@@XMSalloc:
        push ds:himembase               ; =linear address of block
        push es                         ; XMSinit destroys es
        call XMSinit
        pop es
        mov edx,ds:himembase
        add edx,ds:code32a              ; edx=phys addr of start XMS block
        or edx,7                        ; User (=page addressable regardless
        pop ds:himembase                ;  the CPL); Read and write; Present
        mov cx,di                       ; cx=first free entry in page table
        add cx,3*4                      ; 3 pages; 1 for page dir, 1 for
@@next_XMSpage:                         ;  server's page, 1 to support 4GB
        mov es:[di],edx
        add edx,4096                    ; Next page resides 4K higher in mem
        add di,4
        cmp di,cx
        jb short @@next_XMSpage         ; edi=first free entry in page tables
        mov VCPIpage_top,edi
                                        ; Set up page directory address (CR3)
@@setup_pagedir:
        mov ebx,ds:VCPIdir_base
        shr ebx,12                      ; number of 4KB page where dir in mem
        mov eax,es:[ebx*4 + 4096]       ; Get page dir address in page table
        and eax,0fffff000h              ; 12 lowest bits reserved by Intel
        mov ds:VCPIcr3,eax
                                        ; Set up page directory
        xor di,di
        mov cx,2                        ; 2 page tbl; 1 from VCPI server and 1
@@next_table:                           ;  for 4MB to store page tbls for 4GB
        inc ebx                         ; Next page directory entry
        mov eax,es:[ebx*4 + 4096]       ; Get address of next page table
        and eax,0fffff000h              ; Clear page attributes
        or eax,7                        ; User (=page addressable regardless
        stosd                           ;  the CPL); Read and write; Present
        loop @@next_table

        call init_callbacks

        mov ax,ds:VCPIsel0
        mov word ptr ds:VCPIcall[4],ax

        mov ax,ds:mainTSSsel
        mov ds:VCPItask,ax

        mov ax,ds:code16sel
        mov word ptr ds:exitPM32[4],ax
        mov ds:exitPM32[0],offset VCPIexit

        ret
;
VCPIinitPM:
                                        ; Copy page dir & 1st 2 page tables
        mov esi,ds:VCPIdir_base         ;  from conv to extended memory
        mov edi,ds:himembase
        add edi,ds:code32a
        mov ds:VCPIdir_base,edi

        mov ecx,VCPIpage_top            ; ecx=amount mem for page dir/tables
        shr ecx,2
        mov ds,cs:zerosel
        mov es,cs:zerosel
        cld
;       db 67h                          ; Address-size prefix -> Move ecx (!)
;       rep movsd                       ;  dwords from ds:esi to es:edi
        rep movs dword ptr es:[edi],ds:[esi]

                                        ; Set up page directory address (CR3)
        mov ds,cs:data16sel
        mov esi,ds:VCPIdir_base
        mov ebx,esi
        shr ebx,12                      ; number of 4KB page where dir in mem
        mov eax,es:[ebx*4 + esi+4096]   ; Get page dir address in page table
        and eax,0fffff000h              ; 12 lowest bits reserved by Intel
        mov ds:VCPIcr3,eax
        mov cr3,eax                     ; Make page dir & page tables active

        mov edi,esi                     ; Set up page dir's 1st 2 entries
        mov ecx,2                       ;  (make entries point to page tables
@@next_table:                           ;  in ext mem instead of conv mem)
        inc ebx                         ; Next page directory entry
        mov eax,es:[ebx*4 + esi+4096]   ; Get address of next page table
        and eax,0fffff000h              ; Clear page attributes
        or eax,7                        ; User (=page addressable regardless
        db 67h                          ;  the CPL); Read and write; Present
        stosd                           ; stosd at es:edi (above 1MB ;)
        loop @@next_table

        cmp ds:XMSmem,0                 ; Check if memory has already been
        jne short @@XMSpages            ;  allocated through XMS -> Make page
                                        ;  tables for it

                                        ; Allocate memory & create page
        mov edi,VCPIpage_top            ;  tables
        add edi,esi                     ; esi=VCPIdir_base
        mov ecx,edi                     ; ecx=first free entry in page table
        add ecx,ds:VCPImem              ; Amount of mem allocated is diff
@@next_page:                            ;  between ecx and edi
        mov ax,0de04h                   ; al=4 -> Allocate a 4KB page
        call fword ptr ds:VCPIcall
        cmp ah,0                        ; No error?
        jne short @@last_page           ; edx=physical page address
        and edx,0fffff000h              ; Clear page attributes
        or edx,7                        ; User (=page addressable regardless
        mov es:[edi],edx                ;  the CPL); Read and write; Present
        add edi,4
        cmp edi,ecx
        jb short @@next_page
@@last_page:                            ; edi=first free entry in page tables
        sub edi,esi                     ; esi=VCPIdir_base
        mov VCPIpage_top,edi

        jmp short setup_pagedir
@@XMSpages:                             ; Create page tables for allocated
        mov edi,VCPIpage_top            ;  XMS memory
        add edi,esi                     ; esi=VCPIdir_base
        mov ecx,ds:XMSmem
        sub ecx,3*4                     ; Already made 3 page table entries   
        add ecx,edi                     ; Amount mem that page table entries
                                        ;  are created for is diff ecx and edi
        mov edx,es:[edi-4]              ; edx=phys addr of last allocated page
        add edx,4096                    ; Next page resides 4K higher in mem
        or edx,7                        ; User (=page addressable regardless
@@next_XMSpage:                         ;  the CPL); Read and write; Present
        mov es:[edi],edx
        add edx,4096                    ; Next page resides 4K higher in mem
        add edi,4
        cmp edi,ecx
        jb short @@next_XMSpage
        sub edi,esi                     ; esi=VCPIdir_base
        mov VCPIpage_top,edi            ; edi=first free entry in page tables

setup_pagedir:
        mov eax,ds:himembase
        add eax,edi                     ; Save page directory and page tables
        mov ds:himembase,eax

        sub edi,4096                    ; edi=first free entry in page tables
        shl edi,10                      ;    =number of used 1KB blocks
        sub edi,ds:code32a              ; =offset of 1st byte above block
        mov ds:himemtop,edi             ;  from start code32

                                        ; Set up page directory
        mov ebx,esi                     ; esi=VCPIdir_base
        shr ebx,12                      ; number of 4KB page where dir in mem
        mov edi,esi
        mov ecx,VCPIpage_top            ; Round amount up to 4MB units (-4K
        dec ecx                         ;  for dir's size; +4K-1 to round);
        shr ecx,12                      ;  dir entries are done at 4MB incr
@@next_table:                           ; ecx=amount mem in 4MB units
        inc ebx                         ; Next page directory entry
        mov eax,es:[ebx*4 + esi+4096]   ; Get address of next page table
        and eax,0fffff000h              ; Clear page attributes
        or eax,7                        ; User (=page addressable regardless
        db 67h                          ;  the CPL); Read and write; Present
        stosd                           ; stosd at es:edi (above 1MB ;)
        loop @@next_table

        mov ax,ds:code32sel
        mov word ptr ds:VCPItargetPM[4],ax
        mov ds:VCPItargetPM[0],offset retfromVCPI_int

        ret
;
check_DPMI:
        mov ax,1687h                    ; Get RM to PM switch entry point
        int 2fh                         ;  (also check if DPMI present)
        cmp ax,0                        ; No error?
        jne short @@no_DPMI
        test bx,1                       ; 32-bit programs are supported?
        mov dx,offset errmsg5
        mov ds:returncode,1
        jz exit16err                    ; DPMI, but not 32-bit
        mov word ptr DPMIcall[0],di     ; di=entry point offset
        mov word ptr DPMIcall[2],es     ; es=entry point segment
        mov DPMImem_hst,si              ; mem needed by DPMI host (in paras)
        jmp short @@DPMI
@@no_DPMI:
        stc
        ret
@@DPMI:
        clc
        ret
;
DPMIinit:
        mov eax,ds:lomembase
        add eax,ds:code32a
        add eax,0fh                     ; Align lomembase on paragraph
        and eax,0fffffff0h              ;  boundary; Needed for
                                        ;  DPMImem_hst_seg
        shr eax,4
        mov DPMImem_hst_seg,ax
        shl eax,4
        sub eax,ds:code32a              ; eax=low memory base
        movzx ebx,DPMImem_hst
        shl ebx,4                       ; ebx=mem needed by host (in bytes) 
        add eax,ebx
        cmp eax,ds:lomemtop
        ja notenoughlowmem
        mov ds:lomembase,eax

        ret
;
DPMIinitPM:                             ; This part can be done in PM only
        mov ds:code16sel,cs             ; Save 16-bit selectors
        mov ds:data16sel,ds
        mov ds:pspsel,es
                                        ; Allocate & initialize descriptors
        xor ax,ax                       ; ax=0 -> Allocate LDT descriptor
        mov cx,4                        ; cx=number of descriptors to allocate
        int 31h
        mov ds:v86r_dx,offset errmsg5
        mov ds:returncode,3
        jc exit16errDPMI

        mov ds:code32sel,ax             ; ax=first selector in array of dsc
        mov ax,3                        ; ax=3 -> Get next sel increment value
        int 31h
        mov bx,ds:code32sel             ; ax=value to add to get to next sel
        add bx,ax                       ; bx=2nd sel
        mov ds:data32sel,bx
        add bx,ax                       ; bx=3rd sel
        mov ds:zerosel,bx
        add bx,ax                       ; bx=4th sel
        mov ds:envsel,bx

        lar cx,ds:code16sel             ; Load access rights byte in cx
        and cx,6000h                    ; Only the DPL bits are needed
        mov ax,0ch                      ; ax=0ch -> Set descriptor
        mov es,ds:data16sel             ; es:edi=pointer to 8 byte dsc buffer
        mov edi,offset ds:code32dsc
        or ds:code32dsc.access,ch       ; Set the DPL bits to the CPL
        mov bx,ds:code32sel             ; bx=selector
        int 31h                         ; Set descriptor
        mov ds:returncode,4
        jc exit16errDPMI

        mov edi,offset ds:data32dsc
        or ds:data32dsc.access,ch       ; Set the DPL bits to the CPL
        mov bx,ds:data32sel             ; bx=selector
        int 31h                         ; Set descriptor
        mov ds:returncode,5
        jc exit16errDPMI

        mov edi,offset ds:zerodsc
        or ds:zerodsc.access,ch         ; Set the DPL bits to the CPL
        mov bx,ds:zerosel               ; bx=selector
        int 31h                         ; Set descriptor
        mov ds:returncode,6
        jc exit16errDPMI

        mov edi,offset ds:envdsc
        or ds:envdsc.access,ch          ; Set the DPL bits to the CPL
        mov bx,ds:envsel                ; bx=selector
        int 31h                         ; Set descriptor
        mov ds:returncode,7
        jc exit16errDPMI

        mov ax,500h                     ; ax=500h -> Get free mem information
        mov edi,offset DPMImeminfo      ; es:edi=sel:offset of 30h buffer
        int 31h
        mov ds:returncode,8
        jc exit16errDPMI
        cmp DPMImeminfo.max_locked,-1
        je short @@novirtualmem

        mov ax,604h                     ; ax=604h -> Get page size
        int 31h                         ; bx:cx=page size in bytes
        mov ds:returncode,9
        jc exit16errDPMI
        shl ebx,16
        mov bx,cx                       ; ebx=page size
        mov edx,ebx
        shr edx,10                      ; edx=page size in KB
        push edx                        ; Save edx ("mul ebx" detroys edx)
        mov eax,DPMImeminfo.max_locked
        mul ebx                         ; eax=DPMImem in bytes
        shr eax,10                      ; DPMImem in KB
ifdef   WINFRIENDLY
        cmp ds:windows,0                ; Allocating all avail mem under Win
        je short @@no_windows           ;  and locking it slows down the
        shr eax,1                       ;  system terribly :(
@@no_windows:
endif
        mov ds:DPMImem,eax
        pop edx
        jmp short @@DPMIalloc
@@novirtualmem:                         ; Server does not support virtual mem
        mov eax,DPMImeminfo.largest_free;  -> only this field is filled in
        shr eax,10                      ; DPMI mem in KB
        mov ds:DPMImem,eax
        mov edx,4                       ; Assume a default page size of 4KB

@@DPMIalloc:
        shl eax,10
        mov cx,ax
        shld ebx,eax,16                 ; bx:cx=size wanted in bytes
        mov ax,501h                     ; ax=501h -> Allocate mem block
        int 31h                         ; bx:cx=linear address of block
        jnc short @@mem_allocated       ; Some DPMI servers (e.g. Windows 95
        sub ds:DPMImem,edx              ;  when virtual mem is turned off)
        mov eax,ds:DPMImem              ;  return an amount of max lockable
        ja short @@DPMIalloc            ;  mem that can't be allocated -> Try
        mov ds:returncode,10            ;  to allocate a smaller block until
        jmp exit16errDPMI               ;  no error

@@mem_allocated:
        mov word ptr DPMIhandle[2],si   ; si:di=memory block handle
        mov word ptr DPMIhandle[0],di

@@lock_mem:
        mov esi,ds:DPMImem              ; bx:cx=linear address of mem to lock
        shl esi,10
        mov di,si                       ; si:di=size of block to lock (bytes)
        shr esi,16
        mov ax,600h                     ; ax=600h -> Lock linear region
        int 31h
        jnc short @@mem_locked          ; Some DPMI servers (e.g. CWSDPMI in
        sub ds:DPMImem,edx              ;  certain cases) return an amount of
        ja short @@lock_mem             ;  max lockable mem that can't be
        mov ds:returncode,11            ;  locked -> Try to lock a smaller
        jmp exit16errDPMI               ;  block until no error

@@mem_locked:
        mov eax,ds:DPMImem
        cmp eax,HIGHMIN
        jb notenoughhighmemDPMI
;       cmp eax,0                       ; Not necessary; If no mem left to
;       je notenoughhighmemDPMI         ;  alloc program already exited. The
        shl eax,10                      ;  same applies to locking mem.
        mov ds:himemtop,eax             ; Now only _size_ of block

                                        ; Resize the allocated memory block
        mov cx,ax                       ;  to the size of the locked block.
        shld ebx,eax,16                 ; bx:cx=new size wanted in bytes
        mov si,word ptr DPMIhandle[2]   ; si:di=handle of block to resize
        mov di,word ptr DPMIhandle[0]
        mov ax,503h                     ; ax=503h -> Resize mem block
        int 31h
        mov ds:returncode,12
        jc exit16errDPMI

        mov word ptr DPMIhandle[2],si   ; si:di=new handle of memory block
        mov word ptr DPMIhandle[0],di
        shl ebx,16                      ; bx:cx=new linear address of block
        mov bx,cx
        sub ebx,ds:code32a
        mov ds:himembase,ebx
        add ds:himemtop,ebx

        mov ax,205h                     ; ax=205h -> Set PM interrupt vector
        mov bl,RMCALL_VECT              ; bl=interrupt number
        mov cx,ds:code32sel             ; cx:edx=addr of exception handler
        mov edx,offset DPMIcall_int
        int 31h
        mov ds:returncode,13
        jc exit16errDPMI

        mov word ptr ds:exitPM32[4],cs
        mov ds:exitPM32[0],offset DPMIexit

        mov ax,ds:data32sel
        mov es,ax
        mov ds,ax
        mov fs,ax
        mov gs,cs:zerosel

        mov ax,400h                     ; ax=400h -> Get DPMI version info
        int 31h
        mov cx,8
        mov al,dh                       ; dh=value of master PIC base int
        mov di,small offset IRQ0_vect
@@next_masterIRQ:
        stosb
        inc al
        loop @@next_masterIRQ
        mov cx,8
        mov al,dl                      ; dl=value of slave PIC base int
        mov di,small offset IRQ8_vect
@@next_slaveIRQ:
        stosb
        inc al
        loop @@next_slaveIRQ

        mov es,cs:pspsel
        ret
;
start16:
        mov ax,code16
        mov ss,ax
        mov sp,RM_STACKSIZE

        mov ah,30h
        int 21h                         ; bug?

        mov cs:psp_seg,es
        mov ax,es:[2ch]
        mov cs:env_seg,ax

        mov ax,es                       ; Save Program Segment Prefix
        mov ds,ax
        xor si,si
        mov ax,cs
        mov es,ax
        mov di,small offset ds:pspbuffer
        mov cx,256/4
        cld
        rep movsd

        mov eax,code16
        mov ds,ax
        shl eax,4                       ; eax=linear address of code16
        mov ds:code16a,eax
        mov eax,code32                  
        shl eax,4                       ; eax=linear address of code32
        mov ds:code32a,eax
        mov ebx,code32end
        shl ebx,4                       ; ebx=linear address of code32end
        sub ebx,eax                     ; ebx=offset 1st byte above end
        mov ds:lomembase,ebx            ;  code32 from start code32
        call check_memory

        cli
        call check_VCPI                 ; First check VCPI (Privilege Level
        jc short @@no_VCPI              ;  0), DPMI last (nearly always PL 3)
        mov ds:systemtype,2
        call VCPIinit
        jmp short @@setup_dsc
@@no_VCPI:
        call check_V86                  ; Check if the processor is already
        jc short @@check_DPMI           ;  in PM -> Must use interface to
        call check_XMS                  ;  enter PM
        jc short @@no_XMS
        mov ds:systemtype,1
        call XMSinit
        jmp short @@setup_dsc
@@no_XMS:
        call check_V86
        jc short @@check_DPMI
        mov ds:systemtype,0
        call rawinit
        jmp short @@setup_dsc
@@check_DPMI:
        mov dx,offset errmsg8
        mov ds:returncode,0ffh          
        call check_DPMI
        jc exit16err                    ; If no DPMI -> No way for this
        mov ds:systemtype,3             ;  program to enter PM
        call DPMIinit

@@setup_dsc:
        mov eax,ds:code32a              ; Set up basic descriptors    
        add dword ptr ds:gdtr[2],eax
        mov ds:code32dsc.base0_15,ax
        mov ds:data32dsc.base0_15,ax
        add ds:mainTSSdsc.base0_15,ax
        shr eax,8
        mov ds:code32dsc.base16_23,ah
        mov ds:data32dsc.base16_23,ah
        add ds:mainTSSdsc.base16_23,ah
        mov eax,ds:code16a
        mov ds:code16dsc.base0_15,ax
        mov ds:data16dsc.base0_15,ax
        shr eax,8
        mov ds:code16dsc.base16_23,ah
        mov ds:data16dsc.base16_23,ah

        movzx eax,ds:psp_seg            ; Set up PSP descriptor
        shl eax,4
        mov ds:pspdsc.base0_15,ax
        shr eax,8
        mov ds:pspdsc.base16_23,ah

        movzx eax,ds:env_seg            ; Set up environment descriptor
        shl eax,4
        mov ds:envdsc.base0_15,ax
        shr eax,8
        mov ds:envdsc.base16_23,ah

        call check_windows              ; present -> windows=1 else windows=0

        mov es,ds:psp_seg               ; es=segment of the block
        mov ebx,ds:code32a
        add ebx,ds:lomembase            ; ebx=offset 1st byte free low mem
        add ebx,0fh                     ; ebx=idem, round up to paragraphs

        mov edx,ebx
        and edx,0fffffff0h              ; Align lomembase on paragraph
        sub edx,ds:code32a              ;  boundary -> MCB at lomembase
        mov ds:lomembase,edx

        shr ebx,4                       ; bx=offset in paragraph units
        sub bx,ds:psp_seg               ; bx=new block size in paragraphs
        mov ah,4ah                      ; ah=4ah -> Modify allocated mem block
        int 21h
        mov dx,offset errmsg0           ; int 21h/ah=4ah places 1 byte + 2
        mov ds:returncode,0ffh          ;  words at 1st para above lomembase
        jc exit16err                    ;  if not already on para boundary.
                                        ;  These bytes are part of the 16
                                        ;  byte MCB and must be preserved.
        add ds:lomembase,16             ; Save MCB

        movzx bx,ds:systemtype
        shl bx,1                        ; Elements of enterPM are words
        jmp enterPM[bx]
;
raw_PM:
        lgdt fword ptr ds:gdtr          ; Switch to protected mode
        lidt fword ptr ds:idtr
        mov eax,cr0
        or al,1                         ; Set Protection Enable bit
        mov cr0,eax
        db 0eah                         ; Exit 16-bit PM, enter 32-bit PM
        dw small offset start32,code32dsc-gdt
;
VCPI_PM:
        mov ax,ds:code16sel
        mov word ptr ds:VCPItargetPM[4],ax
        mov ds:VCPItargetPM[0],offset @@inPM
        mov esi,offset VCPIcr3          ; Switch to protected mode
        add esi,ds:code32a              ; esi=system register table address
        mov ax,0de0ch                   ; al=ch -> Switch CPU mode
        int 67h
@@inPM:
        mov ax,cs:data16sel
        mov ds,ax
        mov es,ax
        mov fs,ax
        mov gs,ax
        mov ss,ax
        mov sp,RM_STACKSIZE
        call VCPIinitPM                 ; Make page tables in extended mem to
                                        ;  _really_ support 4GB
        db 0eah                         ; Exit 16-bit PM, enter 32-bit PM
        dw small offset start32,code32dsc-gdt
;
DPMI_PM:
        mov es,DPMImem_hst_seg
        mov ax,1                        ; Indicate 32-bit application
        call DPMIcall                   ; Switch to protected mode
        mov dx,offset errmsg5
        mov ds:returncode,2
        jc exit16err                    ; Error, still in real mode
        call DPMIinitPM                 ; Init that can be done in PM only

        mov ax,cs:data32sel
        mov ss,ax
        mov esp,PM_STACKSIZE

        mov word ptr [esp-4],small offset DPMIstart32
        mov ax,cs:code32sel
        mov [esp-2],ax
        jmp dword ptr [esp-4]           ; Exit 16-bit PM, enter 32-bit PM
;
RM_int:                                 ; Call real mode interrupt handler
        lidt fword ptr cs:idtrRM        ; Switch to real mode
        mov eax,cr0
        and al,0feh                     ; Clear Protection Enable bit
        mov cr0,eax
        db 0eah                         ; JMP intersegment (sets cs) opcode
        dw $+4,code16                   ; immediate target address of JMP

        mov di,offset @@toPM
        mov bh,0                        ; necessary for RM_IRQ
        mov bl,cs:RM_intnum
        cmp bl,8                        ; int<8?
        jb short @@noIRQ
        sub bl,8                        ; necessary for RM_IRQ
        cmp bl,7                        ; int>=8 && int<=15?
        jbe RM_IRQ
        cmp bl,70h-8                    ; int>15 && int<70h?
        jb short @@noIRQ
        sub bl,70h-16                   ; necessary for RM_IRQ
        cmp bl,8+7                      ; int>=70h && int<=77h?
        jbe RM_IRQ
@@noIRQ:
        mov ax,code16
        mov ss,ax
        mov esp,RM_STACKSIZE            ; Must be _e_sp ('flat real mode'
        mov ax,cs:v86r_flags            ;  effect (hidden part of seg regs))
        and ah,not 2                    ; Keep interrupts disabled
        push ax
        popf
        mov ds,cs:v86r_ds
        mov es,cs:v86r_es
        mov fs,cs:v86r_fs
        mov gs,cs:v86r_gs
        mov eax,cs:v86r_eax
        mov ebx,cs:v86r_ebx
        mov ecx,cs:v86r_ecx
        mov edx,cs:v86r_edx
        mov esi,cs:v86r_esi
        mov edi,cs:v86r_edi
        mov ebp,cs:v86r_ebp               
        db 0cdh                         ; INT opcode
RM_intnum       db      ?
        mov cs:v86r_eax,eax
        pushf
        pop ax
        and ah,not 42h                  ; Clear NT flag and IF (Int Enable);
        push ax                         ;  int 21h/ah=4bh sets these flags
        popf
        mov cs:v86r_flags,ax
        mov cs:v86r_ebx,ebx
        mov cs:v86r_ecx,ecx
        mov cs:v86r_edx,edx
        mov cs:v86r_esi,esi
        mov cs:v86r_edi,edi
        mov cs:v86r_ebp,ebp
        mov cs:v86r_ds,ds
        mov cs:v86r_es,es
        mov cs:v86r_fs,fs
        mov cs:v86r_gs,gs

@@toPM:
        lgdt fword ptr cs:gdtr          ; Switch to protected mode
        lidt fword ptr cs:idtr
        mov eax,cr0
        or al,1                         ; Set Protection Enable bit
        mov cr0,eax
        db 0eah
        dw small offset retfromRM_int,code32dsc-gdt
;
VCPI_int:                               ; Call real mode interrupt handler
        mov di,offset @@toPM
        mov bh,0                        ; necessary for RM_IRQ
        mov bl,cs:VCPI_intnum
        cmp bl,8                        ; int<8?
        jb short @@noIRQ
        sub bl,8                        ; necessary for RM_IRQ
        cmp bl,7                        ; int>=8 && int<=15?
        jbe RM_IRQ
        cmp bl,70h-8                    ; int>15 && int<70h?
        jb short @@noIRQ
        sub bl,70h-16                   ; necessary for RM_IRQ
        cmp bl,8+7                      ; int>=70h && int<=77h?
        jbe RM_IRQ
@@noIRQ:
        mov ax,cs:v86r_flags            ; VCPI server doesn't do this
        and ah,not 2                    ; Keep interrupts disabled
        push ax
        popf
        mov eax,cs:v86r_eax
        mov ebx,cs:v86r_ebx
        mov ecx,cs:v86r_ecx
        mov edx,cs:v86r_edx
        mov esi,cs:v86r_esi
        mov edi,cs:v86r_edi
        mov ebp,cs:v86r_ebp               
        db 0cdh                         ; INT opcode
VCPI_intnum     db      ?
        cli
        mov cs:v86r_eax,eax
        mov cs:v86r_ebx,ebx
        mov cs:v86r_ecx,ecx
        mov cs:v86r_edx,edx
        mov cs:v86r_esi,esi
        mov cs:v86r_edi,edi
        mov cs:v86r_ebp,ebp
        mov cs:v86r_ds,ds
        mov cs:v86r_es,es
        mov cs:v86r_fs,fs
        mov cs:v86r_gs,gs
        pushf
        pop ax
        mov cs:v86r_flags,ax

@@toPM:
        mov ax,cs:code32sel
        mov word ptr cs:VCPItargetPM[4],ax
        mov cs:VCPItargetPM[0],offset retfromVCPI_int
        mov esi,offset VCPIcr3          ; Switch to protected mode
        add esi,cs:code32a              ; esi=system register table address
        mov ax,0de0ch                   ; al=ch -> Switch CPU mode
        int 67h                         ; code32sel:retfromVCPI_int
;
RM_IRQ:                                 ; Call real mode IRQ handler
        lss sp,cs:stackptrRM
        sub cs:spRM,RM_STACKSIZE/4
        push 3002h                      ; IRQ handler is CALLed but does IRET
        shl bx,2
        call dword ptr cs:[orgIVT+bx]   ; Call original RM handler
        add cs:spRM,RM_STACKSIZE/4      ; Return to exiting code in RM_int /
        jmp di                          ;  VCPI_int when IRQ handler returns
;
ifdef   RMCB_ON
RMcallbackIRQ0:                         ; Reflect an IRQ in RM back to PM
        push bx
        mov bh,0
        mov bl,cs:IRQ0_vect
        jmp RMcallbackIRQ
RMcallbackIRQ1:
        push bx
        mov bh,0
        mov bl,cs:IRQ1_vect
        jmp RMcallbackIRQ
RMcallbackIRQ2:
        push bx
        mov bh,0
        mov bl,cs:IRQ2_vect
        jmp RMcallbackIRQ
RMcallbackIRQ3:
        push bx
        mov bh,0
        mov bl,cs:IRQ3_vect
        jmp RMcallbackIRQ
RMcallbackIRQ4:
        push bx
        mov bh,0
        mov bl,cs:IRQ4_vect
        jmp RMcallbackIRQ
RMcallbackIRQ5:
        push bx
        mov bh,0
        mov bl,cs:IRQ5_vect
        jmp RMcallbackIRQ
RMcallbackIRQ6:
        push bx
        mov bh,0
        mov bl,cs:IRQ6_vect
        jmp RMcallbackIRQ
RMcallbackIRQ7:
        push bx
        mov bh,0
        mov bl,cs:IRQ7_vect
        jmp RMcallbackIRQ
RMcallbackIRQ8:
        push bx
        mov bh,0
        mov bl,cs:IRQ8_vect
        jmp RMcallbackIRQ
RMcallbackIRQ9:
        push bx
        mov bh,0
        mov bl,cs:IRQ9_vect
        jmp RMcallbackIRQ
RMcallbackIRQ10:
        push bx
        mov bh,0
        mov bl,cs:IRQ10_vect
        jmp RMcallbackIRQ
RMcallbackIRQ11:
        push bx
        mov bh,0
        mov bl,cs:IRQ11_vect
        jmp RMcallbackIRQ
RMcallbackIRQ12:
        push bx
        mov bh,0
        mov bl,cs:IRQ12_vect
        jmp RMcallbackIRQ
RMcallbackIRQ13:
        push bx
        mov bh,0
        mov bl,cs:IRQ13_vect
        jmp RMcallbackIRQ
RMcallbackIRQ14:
        push bx
        mov bh,0
        mov bl,cs:IRQ14_vect
        jmp RMcallbackIRQ
RMcallbackIRQ15:
        push bx
        mov bh,0
        mov bl,cs:IRQ15_vect
RMcallbackIRQ:
        push eax edx esi edi cx ds es fs gs
        mov cx,ss
        mov edx,esp

        mov eax,cs:idt
        add eax,cs:code32a
        shr eax,4
        mov ds,ax                       ; ds=segment of IDT
        shl bx,3                        ; bx=offset of dsc in IDT
        mov di,[bx+6]
        shl edi,16
        mov di,[bx]                     ; edi=offset of PM IRQ handler
        mov bx,[bx+2]                   ; bx=selector of PM IRQ handler

        cmp cs:systemtype,2
        je short @@VCPItoPM
        lgdt fword ptr cs:gdtr          ; Switch to protected mode
        lidt fword ptr cs:idtr
        mov eax,cr0
        or al,1                         ; Set Protection Enable bit
        mov cr0,eax
        db 0eah                         ; JMP intersegment (sets cs) opcode
        dw @@inPM,code16dsc-gdt         ; immediate target address of JMP

@@VCPItoPM:
        mov ax,cs:code16sel
        mov word ptr cs:VCPItargetPM[4],ax
        mov cs:VCPItargetPM[0],offset @@inPM
        mov esi,offset VCPIcr3          ; Switch to protected mode
        add esi,cs:code32a              ; esi=system register table address
        mov ax,0de0ch                   ; al=ch -> Switch CPU mode
        int 67h

@@inPM:
        mov ax,cs:data32sel
        mov ds,ax                       ; seg regs must have a valid PM value
        mov es,ax
        mov fs,ax
        mov gs,ax

        assume ds:code32

        sub espPM4,PM_STACKSIZE/4
        lss esp,stackptrPM4
        add stackoff,stackptrPM2-stackptrPM
        push large 3002h                ; 32-bit stack -> IRETD pops 12 bytes
        mov [esp-6],edi
        mov [esp-2],bx
        call fword ptr [esp-6]          ; Call PM IRQ handler

        sub stackoff,stackptrPM2-stackptrPM
        add espPM4,PM_STACKSIZE/4
        cmp systemtype,2
        je short @@VCPItoRM
        lidt fword ptr cs:idtrRM        ; Switch to real mode
        mov eax,cr0
        and al,0feh                     ; Clear Protection Enable bit
        mov cr0,eax
        db 0eah
        dw @@inRM,code16

@@VCPItoRM:                             ; ss:esp=sel:off system register tbl
        push large 0                    ; V86 gs
        push large 0                    ; V86 fs
        push large code16               ; V86 ds
        push large code16               ; V86 es
        push ecx                        ; V86 ss
        push edx                        ; V86 esp
        push large 3002h                ; eflags (dummy, _not_ value in V86)
        push large code16               ; V86 cs
        push large offset @@inRM        ; V86 eip
        mov ds,VCPIsel1

        assume ds:code16

        mov ax,0de0ch                   ; al=ch -> Switch CPU mode
        call fword ptr cs:VCPIcall      ; Switch to V86 mode

@@inRM:
        mov ss,cx                       ; Restore stack active on entry
        mov esp,edx
        pop gs fs es ds cx edi esi edx eax bx
        iret
endif
;
rawexit:                            
        lidt fword ptr idtrRM           ; Switch to real mode
        mov eax,cr0
        and al,0feh                     ; Clear Protection Enable bit
        mov cr0,eax
        db 0eah                         ; JMP intersegment (sets cs) opcode
        dw $+4,code16                   ; immediate target address of JMP
        mov ax,cs                       ; back in real mode
        mov ds,ax
        mov es,ax
        mov ss,ax
        mov sp,RM_STACKSIZE
        call restoreIVT
        call disableA20
        mov al,ds:returncode
        mov ah,4ch
        int 21h
;
XMSexit:
        lidt fword ptr idtrRM           ; Switch to real mode
        mov eax,cr0
        and al,0feh                     ; Clear Protection Enable bit
        mov cr0,eax
        db 0eah                         ; JMP intersegment (sets cs) opcode
        dw $+4,code16                   ; immediate target address of JMP
        mov ax,cs                       ; back in real mode
        mov ds,ax
        mov es,ax
        mov ss,ax
        mov sp,RM_STACKSIZE
XMSdealloc:
        call restoreIVT
        push word ptr ds:returncode     ; Save returncode for when no error
        mov dx,XMShandle                ; dx=block handle
        mov ah,0dh                      ; ah=0dh -> Unlock extended mem block
        call XMScall
        cmp al,0                        ; Error?
        mov dx,offset errmsg7
        mov ds:returncode,4
        je exit16err

        mov dx,XMShandle                ; dx=block handle
        mov ah,0ah                      ; ah=0ah -> Release ext mem block
        call XMScall
        cmp al,0                        ; Error?
        mov dx,offset errmsg7
        mov ds:returncode,5
        je exit16err

        mov ah,6                        ; ah=6 -> Local disable A20 line
        call XMScall
        cmp al,0                        ; Error?
        mov dx,offset errmsg4
        mov ds:returncode,0ffh
        je exit16err

        pop ax                          ; al=returncode
        mov ah,4ch
        int 21h
;
VCPIexit:                               ; This code runs in 16-bit PM
        mov es,ds:zerosel               ; Deallocate pages
        mov edi,VCPIpage_base
        mov ecx,VCPIpage_top
        sub ecx,edi
        jbe short @@dealloc_done
        shr ecx,2                       ; ecx=number of pages to free
        add edi,ds:VCPIdir_base
@@next_page:
        mov edx,es:[edi]                ; edx=physical page address
        and edx,0fffff000h
        mov ax,0de05h                   ; al=5 -> Free a 4KB page
        call fword ptr cs:VCPIcall
        add edi,4
        loop @@next_page
@@dealloc_done:
                                        ; ss:esp=sel:off system register tbl
        push large 0                    ; V86 gs
        push large 0                    ; V86 fs
        push large code16               ; V86 ds
        push large code16               ; V86 es
        push large code16               ; V86 ss
        push large RM_STACKSIZE         ; V86 esp
        push large 3002h                ; eflags (dummy, _not_ value in V86)
        push large code16               ; V86 cs
        push large offset @@inRM        ; V86 eip
        mov ds,cs:VCPIsel1
        mov ax,0de0ch                   ; al=ch -> Switch CPU mode
        call fword ptr cs:VCPIcall      ; Switch to V86 mode
@@inRM:                                 ; back in V86 mode
        call check_XMS                  ; If XMS was present at startup mem
        jnc XMSdealloc                  ;  has been allocated through XMS, so
        call restoreIVT                 ;  deallocation through XMS too
        mov al,ds:returncode
        mov ah,4ch
        int 21h
;
DPMIexit:                               ; This code runs in 16-bit PM
        mov al,ds:returncode
        mov ah,4ch
        int 21h

;
exit16err:                              ; Exit program with error message
        mov ah,9                        ; ah=9 -> Print string at ds:dx
        int 21h
        cmp ds:returncode,0ffh          ; If returncode=ffh -> Don't print it
        je short @@exit

        mov ah,6                        ; ah=6 -> Direct console I/O
        movzx si,ds:returncode
        shr si,4                        ; Print most significant nibble 1st
        mov dl,ds:hextbl[si]            ; dl=character to output
        int 21h

        movzx si,ds:returncode          ; Print least significant nibble 2nd
        and si,0fh
        mov dl,ds:hextbl[si]
        int 21h

@@exit:
        mov ah,0                        ; ah=0 -> Wait for key and read char
        int 16h

        mov al,ds:returncode
        mov ah,4ch
        int 21h
;
exit16errDPMI:                          ; Exit program with error message
        mov ds:v86r_ds,code16           ; Real mode ds is needed
        mov ds:v86r_ah,9                ; ah=9 -> Print string at ds:dx
        mov ds:v86r_ss,0                ; Let DPMI host provide a stack
        mov ds:v86r_sp,0
        mov ax,300h                     ; ax=300h -> Simulate real mode int
        mov cx,ds                
        mov es,cx                       ; cx=number of words to copy from
        xor cx,cx                       ;  protected mode to real mode stack
        mov edi,offset ds:v86r_edi      ; es:edi=address of RM call structure
        mov bh,0                        ; bh=flags
        mov bl,21h                      ; bl=interrupt number
        int 31h
        cmp ds:returncode,0ffh          ; If returncode=ffh -> Don't print it
        je short @@exit

        mov ds:v86r_ah,6                ; ah=6 -> Direct console I/O
        movzx si,ds:returncode
        shr si,4                        ; Print most significant nibble 1st
        mov si,word ptr ds:hextbl[si]
        mov ds:v86r_dx,si               ; dl=character to output
        int 31h

        movzx si,ds:returncode          ; Print least significant nibble 2nd
        and si,0fh
        mov si,word ptr ds:hextbl[si]
        mov ds:v86r_dx,si
        int 31h

@@exit:
        mov ds:v86r_ah,0                ; ah=0 -> Wait for key and read char 
        mov bl,16h
        int 31h

        mov al,ds:returncode
        mov ah,4ch
        int 21h

code16  ends

code32  segment para public use32
        assume cs:code32, ds:code32, ss:code32

;32-bit data
callbackoff     label                   ; temporary init variables
ifdef   RMCB_ON
dw      RMcallbackIRQ0, RMcallbackIRQ1, RMcallbackIRQ2, RMcallbackIRQ3
dw      RMcallbackIRQ4, RMcallbackIRQ5, RMcallbackIRQ6, RMcallbackIRQ7
dw      RMcallbackIRQ8, RMcallbackIRQ9, RMcallbackIRQ10, RMcallbackIRQ11
dw      RMcallbackIRQ12, RMcallbackIRQ13, RMcallbackIRQ14, RMcallbackIRQ15
endif
excoff          label
dw      small exc0, small exc1, small exc2, small exc3, small exc4
dw      small exc5, small exc6, small exc7, small exc8, small exc9
dw      small exc10, small exc11, small exc12, small exc13, small exc14
dw      small exc15, small exc16
IRQoff          label
dw      small IRQ0, small IRQ1, small IRQ2, small IRQ3, small IRQ4
dw      small IRQ5, small IRQ6, small IRQ7, small IRQ8, small IRQ9
dw      small IRQ10, small IRQ11, small IRQ12, small IRQ13, small IRQ14
dw      small IRQ15
intoff          label
dw      small int8, small int9, small int10, small int11, small int12
dw      small int13, small int14, small int15
if      $-code32 gt PM_STACKSIZE          ; Stack large enough for temp vars?
err     "Increase value of PM_STACKSIZE"
endif

org PM_STACKSIZE                        ; protected mode stack
gdt             seg_descriptor  <>
code32dsc       seg_descriptor  <0ffffh, 0, 0, 10011010b, 11001111b, 0>
data32dsc       seg_descriptor  <0ffffh, 0, 0, 10010010b, 11001111b, 0>
zerodsc         seg_descriptor  <0ffffh, 0, 0, 10010010b, 11001111b, 0>
code16dsc       seg_descriptor  <0ffffh, 0, 0, 10011010b, 0,         0>
data16dsc       seg_descriptor  <0ffffh, 0, 0, 10010010b, 0,         0>
pspdsc          seg_descriptor  <0ffh,   0, 0, 10010010b, 0,         0>
envdsc          seg_descriptor  <0ffffh, 0, 0, 10010010b, 0,         0>
mainTSSdsc      seg_descriptor  <2067h, small mainTSS, 0, 10001001b, 0, 0>
VCPIdsc0        seg_descriptor  <>      ; 68h bytes for task state + 2000h
VCPIdsc1        seg_descriptor  <>      ;  bytes for I/O map
VCPIdsc2        seg_descriptor  <>
freedsc         seg_descriptor  N_DSC dup(<>)
gdtr            dw      $-gdt-1, small gdt, 0

align 4
code32sel       dw      code32dsc  -gdt
data32sel       dw      data32dsc  -gdt
zerosel         dw      zerodsc    -gdt
code16sel       dw      code16dsc  -gdt
data16sel       dw      data16dsc  -gdt
pspsel          dw      pspdsc     -gdt
envsel          dw      envdsc     -gdt
mainTSSsel      dw      mainTSSdsc -gdt
VCPIsel0        dw      VCPIdsc0   -gdt
VCPIsel1        dw      VCPIdsc1   -gdt
VCPIsel2        dw      VCPIdsc2   -gdt
freesel         dw      freedsc    -gdt ; selector base of allocation

idt             dd      0               ; offset of IDT from code32
idtr            dw      0,0,0
mainTSS         TSS     <, PM_STACKSIZE, data32dsc-gdt>
                db      2000h dup(0)    ; 1 bit for each port; 64K / 8 = 8K

IRQhandler      label
IRQ0handler     dd IRQ0
                dw code32dsc-gdt
IRQ1handler     dd IRQ1
                dw code32dsc-gdt
IRQ2handler     dd IRQ2
                dw code32dsc-gdt
IRQ3handler     dd IRQ3
                dw code32dsc-gdt
IRQ4handler     dd IRQ4
                dw code32dsc-gdt
IRQ5handler     dd IRQ5
                dw code32dsc-gdt
IRQ6handler     dd IRQ6
                dw code32dsc-gdt
IRQ7handler     dd IRQ7
                dw code32dsc-gdt

exchandler      label
exc8handler     dd exc8
                dw code32dsc-gdt
exc9handler     dd exc9
                dw code32dsc-gdt
exc10handler    dd exc10
                dw code32dsc-gdt
exc11handler    dd exc11
                dw code32dsc-gdt
exc12handler    dd exc12
                dw code32dsc-gdt
exc13handler    dd exc13
                dw code32dsc-gdt
exc14handler    dd exc14
                dw code32dsc-gdt
exc15handler    dd exc15
                dw code32dsc-gdt

align 4
mPICbase        label   byte            ; master PIC base interrupt
IRQ0_vect       db      8
IRQ1_vect       db      9
IRQ2_vect       db      0ah
IRQ3_vect       db      0bh
IRQ4_vect       db      0ch
IRQ5_vect       db      0dh
IRQ6_vect       db      0eh
IRQ7_vect       db      0fh
sPICbase        label   byte            ; slave PIC base interrupt
IRQ8_vect       db      70h
IRQ9_vect       db      71h
IRQ10_vect      db      72h
IRQ11_vect      db      73h
IRQ12_vect      db      74h
IRQ13_vect      db      75h
IRQ14_vect      db      76h
IRQ15_vect      db      77h

v86r_edi        label   dword
v86r_di         dw      0
                dw      0
v86r_esi        label   dword
v86r_si         dw      0
                dw      0
v86r_ebp        label   dword
v86r_bp         dw      0
                dw      0
                dd      0               ; Reserved under DPMI
v86r_ebx        label   dword
v86r_bx         label   word
v86r_bl         db      0
v86r_bh         db      0
                dw      0
v86r_edx        label   dword
v86r_dx         label   word
v86r_dl         db      0
v86r_dh         db      0
                dw      0
v86r_ecx        label   dword
v86r_cx         label   word
v86r_cl         db      0
v86r_ch         db      0
                dw      0
v86r_eax        label   dword
v86r_ax         label   word
v86r_al         db      0
v86r_ah         db      0
                dw      0
v86r_flags      dw      0
v86r_es         dw      0
v86r_ds         dw      0
v86r_fs         dw      0
v86r_gs         dw      0
v86r_ip         dw      0               ; Reserved under DPMI, but ignored
v86r_cs         dw      0               ; Reserved under DPMI, but ignored
v86r_sp         dw      0
v86r_ss         dw      0

DPMImem         dd      0               ; mem available through DPMI in KB

; system register table start 
align 4
VCPIcr3         dd      0               ; physical address of page directory
VCPIgdtr        dd      gdtr
VCPIidtr        dd      idtr
VCPIldt         dw      0
VCPItask        dw      0
VCPItargetPM    dd      0               ; 32-bit offset
                dw      0               ; selector
; system register table end 

VCPIdir_base    dd      0               ; linear address of page directory
VCPIcall        dd      0               ; 32-bit offset
                dw      0               ; selector
VCPImem         dd      0               ; mem available through VCPI in KB
XMSmem          dd      0               ; mem available through XMS in KB
systemtype      db      0               ; 0=raw, 1=XMS, 2=VCPI, 3=DPMI

pspbuffer       db      256 dup(0)
psp_seg         dw      0
env_seg         dw      0
paramblock      dw      0, small cmdtail, code32, 5ch, 0, 6ch, 0
cmdtail         db      0               ; command tail counter
                db      '/c '           ; COMMAND.COM: execute command and
                presettail=($-cmdtail)-1;  return (no space before /c needed)
                db      127-presettail dup(0)
comspec         db      'COMSPEC='

align 4
exitPM32        dd      0               ; 32-bit offset
                dw      0               ; selector
code16a         dd      0               ; linear address of code16
code32a         dd      0               ; linear address of code32
lomembase       dd      0               ; low memory base of allocation
lomemtop        dd      0               ; top of low memory
himembase       dd      0               ; high memory base of allocation
himemtop        dd      0               ; top of high memory
stackptrPM      label   fword
espPM           dd      0               ; temporary variables to save stack
ssPM            dw      0               ;  registers (when calling a RM int)
stackptrPM2     label   fword
espPM2          dd      0               ; necessary if RMCB AND PM IRQ
ssPM2           dw      0               ;  handler calls RM handler
stackptrPM3     label   fword
espPM3          dd      0               ; necessary if nested RMCB
ssPM3           dw      0
stackptrPM4     label   fword
espPM4          dd      PM_STACKSIZE    ; stack used when RMCB switches to PM
ssPM4           dw      data32dsc-gdt
stackoff        dd      0
mPICmask        db      0               ; original master PIC mask
sPICmask        db      0               ; original slave PIC mask
totalextmem     dd      0               ; total extended memory present
hextbl          db      '0123456789ABCDEF'
windows         db      0               ; 0=Windows not active; 1=Windows
returncode      db      0               ; DOS return code

;32-bit code
RMcall_int:                             ; al=RM interrupt number to call
        pushad
        push ds es fs gs
        mov es,cs:data16sel             ; Use es, by default < 64KB (pspsel)

        assume es:code16

        mov es:RM_intnum,al
        mov ebx,cs:stackoff             ; necessary if RMCB is being handled
        mov es:ssPM[ebx],ss             ; Save protected mode ss:esp
        mov es:espPM[ebx],esp

        assume es:nothing

        db 0eah
        dw RM_int,0,code16dsc-gdt
retfromRM_int:                            
        mov ebx,cs:stackoff
        lss esp,cs:stackptrPM[ebx]      ; Restore protected mode ss:esp
        pop gs fs es ds
        popad
ifdef   RMCB_ON
        and byte ptr [esp+9],not 40h    ; Some DOS/BIOS functions set NT flag
endif                                   ;  (only necessary if RMCBs active)
        iretd
;
VCPIcall_int:                           ; al=RM interrupt number to call
        pushad
        push ds es fs gs
        mov ds,cs:data16sel

        assume ds:code16

        mov VCPI_intnum,al
        mov ebx,cs:stackoff             ; necessary if RMCB is being handled
        mov code16:ssPM[ebx],ss         ; Save protected mode ss:esp
        mov code16:espPM[ebx],esp

        assume ds:code32
                                        ; ss:esp=sel:off system register tbl
        push dword ptr cs:v86r_gs       ; V86 gs
        push dword ptr cs:v86r_fs       ; V86 fs
        push dword ptr cs:v86r_ds       ; V86 ds
        push dword ptr cs:v86r_es       ; V86 es
        push large code16               ; V86 ss
        push large RM_STACKSIZE         ; V86 esp
        push 3002h                      ; eflags (dummy, _not_ value in V86)
        push large code16               ; V86 cs
        push large offset VCPI_int      ; V86 eip
        mov ds,cs:VCPIsel1
        mov ax,0de0ch                   ; al=ch -> Switch CPU mode
        call fword ptr cs:VCPIcall      ; Switch to V86 mode
retfromVCPI_int:
        mov ebx,cs:stackoff
        lss esp,cs:stackptrPM[ebx]      ; Restore protected mode ss:esp
        pop gs fs es ds
        popad
        iretd
;
DPMIcall_int:                           ; al=RM interrupt number to call
        push eax ebx ecx edi es
        mov es,cs:data32sel
        mov es:v86r_ss,code16
        mov es:v86r_sp,RM_STACKSIZE
        mov bh,0                        ; bh=flags
        mov bl,al                       ; bl=interrupt number
        mov ax,300h                     ; ax=300h -> Simulate real mode int
        xor cx,cx                       ; cx=number of words to copy from
        mov edi,offset v86r_edi         ;  protected mode to real mode stack
        int 31h                         ; es:edi=address of RM call structure
        pop es edi ecx ebx eax
        iretd
;Exceptions
exc0:                                   ; divide error
        push eax
        mov al,0
        jmp IRQ                         ; Use real mode handler
exc1:                                   ; debug exception
        push eax
        mov al,1
        jmp IRQ                         ; Use real mode handler
exc2:                                   ; nonmaskable interrupt
        push eax
        mov al,2
        jmp IRQ                         ; Use real mode handler
exc3:                                   ; breakpoint
        push eax
        mov al,3
        jmp IRQ                         ; Use real mode handler
exc4:                                   ; overflow
        push eax
        mov al,4
        jmp IRQ                         ; Use real mode handler
exc5:                                   ; bounds check
        push eax
        mov al,5
        jmp IRQ                         ; Use real mode handler
exc6:                                   ; invalid opcode
        pushad
        mov dl,6
        jmp exc
exc7:                                   ; coprocessor not available
        push eax
        mov al,7
        jmp IRQ                         ; Use real mode handler
exc8:                                   ; double fault
        pushad
        mov dl,8
        jmp exc
exc9:                                   ; coprocessor segment overrun
        pushad
        mov dl,9
        jmp exc
exc10:                                  ; invalid TSS
        pushad
        mov dl,10
        jmp exc
exc11:                                  ; segment not present
        pushad
        mov dl,11
        jmp exc
exc12:                                  ; stack exception
        pushad
        mov dl,12
        jmp exc
exc13:                                  ; general protection
        pushad
        mov dl,13
        jmp exc
exc14:                                  ; page fault
        cmp systemtype,2                ; This makes sense only under VCPI
        jne short @@exit                ;  (due to usage of 'VCPIdir_base')
        sub esp,4*4                     ; Reserve space for CR2, 2 page table
        push eax ebx                    ;  entries and page dir entr on stack.
        mov es,cs:zerosel               ;  Note that this code puts incorrect
        mov eax,cr2                     ;  values for newly mapped addresses.
        and eax,0fffff000h              ; Discard offset part of linear addr
        shr eax,10
        mov ebx,cs:VCPIdir_base
        sub eax,4
        mov eax,es:[ebx+4096+eax]       ; eax=page table entry before page
        mov [esp+2*4],eax               ;  table entry for lin addr
        mov eax,cr2
        and eax,0fffff000h              ; Discard offset part of linear addr
        shr eax,10
        mov eax,es:[ebx+4096+eax]       ; eax=page table entry for lin addr
        mov [esp+3*4],eax
        mov eax,cr2
        and eax,0ffc00000h              ; Discard page and offset part of
        shr eax,20                      ;  linear addr
        mov eax,es:[ebx+eax]            ; eax=page dir entry for lin addr
        mov [esp+4*4],eax
        mov eax,cr2                     ; CR2 contains the linear address
        mov [esp+5*4],eax               ;  that caused the exception
        pop ebx eax                     ; Restore used eax and ebx
@@exit:
        pushad
        mov dl,14
        jmp exc
exc16:                                  ; coprocessor error
        pushad
        mov dl,16
        jmp exc
exc15:
unexp:                                  ; unexpected interrupt
        pushad
        mov dl,0ffh
exc:
        cld
        mov ax,cs:data32sel
        mov ds,ax
        mov es,ax
        mov fs,ax
        mov ax,zerosel
        mov gs,ax
        mov al,0ffh
        out 21h,al                      ; Mask off all low IRQs
        out 0a1h,al                     ; Mask off all high IRQs
        mov v86r_ax,3                   ; Set video mode 3 (text 80x25 color)
        mov al,10h
        int RMCALL_VECT
        mov edi,0b8000h+2*160           ; Start printing on 3rd row
        sub edi,code32a
        mov ebx,offset hextbl
        mov ecx,2                       ; Print 2 characters
        shl edx,24                      ; exc number in bits 24-31
        mov ah,0fh                      ; bright white
        call putnum                     ; Print exception number
        add edi,152                     ; Goto next row
        mov ah,9                        ; bright blue
        mov ebp,8                       ; Print edi,esi,ebp,esp,ebx,edx,ecx,
putreg:                                 ;  eax = 8 registers
        pop edx                         ; Get register content from stack
        mov cl,8                        ; Print 8 nibbles = 32 bits
        call putnum
        dec ebp
        jnz putreg

        mov ebp,esp                     ; Do a stack dump
        mov esi,esp
        add esi,16*8*4                  ; Print 16 rows of 8 dwords=512 bytes
        mov ah,7                        ; white
putstack:
        mov edx,[ebp]                   ; Get dword from stack
        add ebp,4
        mov cl,8                        ; Print 8 nibbles = 32 bits
        call putnum
        cmp ebp,esi                     ; Last dword on stack printed?
        jb putstack                     ; esp < ebp -> Print another dword
        jmp exit
;
putnum:
        rol edx,4
        mov al,dl
        and al,0fh
        xlat                            ; mov al,[ds:ebx+al]
        stosw                           ; mov [es:edi],ax
        loop putnum
        mov al,' '
        stosw                           ; Print 2 spaces
        stosw
        ret
;IRQs
int8:
        checkifexc 8
        jmp fword ptr cs:IRQ0handler
IRQ0:
        push eax
        mov al,8
        jmp IRQ
int9:
        checkifexc 9
        jmp fword ptr cs:IRQ1handler
IRQ1:
        push eax
        mov al,9
        jmp IRQ
int10:
        checkifexc 10
        jmp fword ptr cs:IRQ2handler
IRQ2:
        push eax
        mov al,10
        jmp IRQ
int11:
        checkifexc 11
        jmp fword ptr cs:IRQ3handler
IRQ3:
        push eax
        mov al,11
        jmp IRQ
int12:
        checkifexc 12
        jmp fword ptr cs:IRQ4handler
IRQ4:
        push eax
        mov al,12
        jmp IRQ
int13:
        checkifexc 13
        jmp fword ptr cs:IRQ5handler
IRQ5:
        push eax
        mov al,13
        jmp short IRQ
int14:
        checkifexc 14
        jmp fword ptr cs:IRQ6handler
IRQ6:
        push eax
        mov al,14
        jmp short IRQ
int15:
        checkifexc 15
        jmp fword ptr cs:IRQ7handler
IRQ7:
        push eax
        mov al,15
        jmp short IRQ
IRQ8:
        push eax
        mov al,70h
        jmp short IRQ
IRQ9:
        push eax
        mov al,71h
        jmp short IRQ
IRQ10:
        push eax
        mov al,72h
        jmp short IRQ
IRQ11:
        push eax
        mov al,73h
        jmp short IRQ
IRQ12:
        push eax
        mov al,74h
        jmp short IRQ
IRQ13:
        push eax
        mov al,75h
        jmp short IRQ
IRQ14:
        push eax
        mov al,76h
        jmp short IRQ
IRQ15:
        push eax
        mov al,77h
IRQ:                                    ; An IRQ in protected mode -> Reflect
        int RMCALL_VECT                 ;  the IRQ to real/V86 mode
        pop eax
        iretd
;
start32:
        mov ax,cs:data32sel
        mov ds,ax
        mov fs,ax
        mov ss,ax
        mov esp,PM_STACKSIZE
        mov es,pspsel
        mov gs,zerosel

        and mainTSSdsc.access,not 2     ; Clear Busy bit (only needed under
        ltr mainTSSsel                  ;  VCPI; TR already loaded by server)

        pushfd
        pop eax
        or ah,30h                       ; bit 12-13 -> IOPL; 3=high, 0=low PL
        and ah,not 40h                  ; Clear Nested Task Flag
        push eax
        popfd

DPMIstart32:                            ; DPMI entry point
        in al,21h                       ; Read Interrupt Mask Register of
        mov mPICmask,al                 ;  master PIC and save it
        and al,0e4h                     ; Enable COM1+COM2 (mouse), keyboard,
        out 21h,al                      ;  timer
        in al,0a1h                      ; Read Interrupt Mask Register of
        mov sPICmask,al                 ;  slave PIC and save it

        mov esi,offset pspbuffer        ; Restore Program Segment Prefix
        xor edi,edi
        mov ecx,256/4
        rep movsd

        mov returncode,0                ; default return code of 0
        sti
        jmp main

;
exit:
        cli
        mov ebx,cs                      ; Check if running at PL 0; CLTS is
        lar ecx,ebx                     ;  a privileged instruction
        and ecx,6000h
        jnz short @@done

        mov al,cs:mPICmask              ; Restore original values in Interrupt
        out 21h,al                      ;  Mask Register of master PIC.
        mov al,cs:sPICmask              ; Restore original values in Interrupt
        out 0a1h,al                     ;  Mask Register of slave PIC.

        clts                            ; Clear Task Switched bit (necessary
                                        ;  for some programs in order to run 
@@done:                                 ;  properly under DOS after this prog)
        mov ax,cs:data16sel
        mov ds,ax
        mov es,ax
        mov fs,ax
        mov gs,ax
        mov ss,ax
        mov esp,RM_STACKSIZE
        jmp fword ptr cs:exitPM32       ; Exit 32-bit PM, enter 16-bit PM


;
; Allocate any memory (first checks low, then high)
; In:
;   ds = data32sel
;   eax = size requested in bytes
; Out:
;   CF=0 -> Memory allocated
;     eax = linear pointer to memory
;     lomembase = lomembase + eax / himembase = himembase + eax
;   CF=1 -> Not enough memory
;     eax = ?
;
getmem:
        push eax
        call getlomem
        jnc short @@exit
        pop eax
        jmp short gethimem
@@exit:
        add esp,4
        ret

;
; Allocate low memory
; In:
;   ds = data32sel
;   eax = size requested in bytes
; Out:
;   CF=0 -> Memory allocated
;     eax = linear pointer to memory
;     lomembase = lomembase + eax
;   CF=1 -> Not enough memory
;     eax = ?
;
getlomem:
        add eax,lomembase
        cmp eax,lomemtop
        ja short @@errexit
        xchg eax,lomembase
        clc
        ret
@@errexit:
        stc
        ret

;
; Allocate high memory
; In:
;   ds = data32sel
;   eax = size requested in bytes
; Out:
;   CF=0 -> Memory allocated
;     eax = linear pointer to memory
;     himembase = himembase + eax
;   CF=1 -> Not enough memory
;     eax = ?
;
gethimem:
        add eax,himembase
        cmp eax,himemtop
        ja short @@errexit
        xchg eax,himembase
        clc
        ret
@@errexit:
        stc
        ret

;
; Get amount of free low memory
; Out:
;   eax = number of bytes free
;
lomemsize:
        mov eax,cs:lomemtop
        sub eax,cs:lomembase
        ret

;
; Get amount of free high memory
; Out:
;   eax = number of bytes free
;
himemsize:
        mov eax,cs:himemtop
        sub eax,cs:himembase
        ret

;
; Allocate descriptor
; In:
;   ds = data32sel
;   cx = number of descriptors to allocate
; Out:
;   CF=0 -> Descriptor successfully allocated
;     ax = first selector in array of descriptors
;   CF=1 -> Not enough free descriptors left
;     ax = ?
;
getdsc:
        push ecx
        cmp systemtype,3                ; Under DPMI the GDT is not part of
        je short @@DPMIgetdsc           ;  this program in memory

        shl cx,3                        ; A descriptor is 8 bytes in size
        add cx,freesel
        mov ax,gdtr                     ; gdtr[0] contains the GDT limit
        inc ax
        cmp cx,ax                       ; Enough descriptors free?
        ja short @@errexit
        mov ax,freesel
        mov freesel,cx
        clc                             ; CF set only on error
        jmp short @@exit

@@DPMIgetdsc:
        xor ax,ax                       ; ax=0 -> Allocate LDT descriptor
        int 31h
        jnc short @@exit                ; ax=first selector in array of dsc

@@errexit:
        stc
@@exit:
        pop ecx
        ret

;
; Set descriptor
; In:
;   ds = data32sel
;   es:edi = pointer to 8 byte descriptor buffer
;   bx = selector
; Out:
;   CF=0 -> Descriptor succesfully set
;   CF=1 -> Descriptor does not reside in the GDT/LDT (non DPMI)
;
setdsc:
        push eax ebx
        cmp systemtype,3                ; Under DPMI the GDT/LDT is not part
        je short @@DPMIsetdsc           ;  of this program in memory

        test ebx,4                      ; Check TI bit
        jnz short @@LDTdsc              ; If TI bit is set -> LDT descriptor
        and ebx,0fffch                  ; Clear RPL bits -> right offset
        movzx eax,gdtr                  ; gdtr[0] contains the GDT limit
        sub eax,7                       ; eax=highest valid GDT entry
        cmp ebx,eax                     ; Check if dsc resides in the GDT
        ja short @@errexit
        mov eax,es:[edi]
        mov dword ptr gdt[ebx],eax
        mov eax,es:[edi+4]
        mov dword ptr gdt[ebx+4],eax
        clc                             ; CF set only on error
        jmp short @@exit

@@LDTdsc:
        and ebx,0fff8h                  ; Clear TI+RPL bits -> right offset
        sldt ax
        and eax,0ffffh                  ; Enable use of _e_ax
        cmp bx,gdt[eax].limit0_15       ; Check if dsc resides in the LDT
        ja short @@errexit
        push edx
        mov dh,gdt[eax].base24_31
        mov dl,gdt[eax].base16_23
        shl edx,16
        mov dx,gdt[eax].base0_15        ; edx=linear address of LDT base
        add ebx,edx                     ; ebx=linear address of dsc in LDT
        pop edx
        sub ebx,code32a
        mov eax,es:[edi]
        mov [ebx],eax
        mov eax,es:[edi+4]
        mov [ebx+4],eax
        clc                             ; CF set only on error
        jmp short @@exit

@@DPMIsetdsc:
        mov ax,0ch                      ; ax=0ch -> Set descriptor
        int 31h
        jnc short @@exit

@@errexit:
        stc
@@exit:
        pop ebx eax
        ret

;
; Get interrupt vector
; In:
;   bl = interrupt number
;   If IDT entries 8-15 are shared between exceptions and IRQs:
;     CF=0 -> Get IRQ handler
;     CF=1 -> Get exception handler
; Out:
;   cx:edx = selector:offset of exception handler
;
getvect:
        pushfd                          ; Needed for CF check
        cmp cs:systemtype,3             ; Under DPMI the IDT is not part of
        je short @@DPMIgetvect          ;  this program in memory

        push ebx
        and ebx,0ffh    
        cmp cs:mPICbase,8               ; Check if IDT entries 8-15 are
        jne short @@notshared           ;  shared
        cmp ebx,8                       ; Check if int is one of the shared
        jb short @@notshared            ;  IDT entries
        cmp ebx,15
        ja short @@notshared

        test byte ptr [esp+4],1         ; Test CF in flags on stack
        jnz short @@getexc              ; CF set -> Get exception handler
        sub ebx,8                       ; ebx=index in IRQhandler
        imul ebx,IRQ1handler-IRQ0handler; ebx=offset from IRQhandler
        mov cx,IRQhandler[ebx+4]
        mov edx,dword ptr IRQhandler[ebx]
        pop ebx
        jmp short @@exit
@@getexc:
        sub ebx,8                       ; ebx=index in exchandler
        imul ebx,exc9handler-exc8handler; ebx=offset from exchandler
        mov cx,exchandler[ebx+4]
        mov edx,dword ptr exchandler[ebx]
        pop ebx
        jmp short @@exit

@@notshared:
        push edi
        mov edi,cs:idt
        shl ebx,3                       ; ebx=offset of int gate dsc in IDT
        add edi,ebx
        mov cx,cs:[edi+2]               ; int gate[2]=selector
        mov dx,cs:[edi+6]               ; int gate[6]=MSW of offset
        shl edx,16
        mov dx,cs:[edi]                 ; int gate[0]=LSW of offset
        pop edi ebx
        jmp short @@exit

@@DPMIgetvect:
        push eax
        mov ax,204h                     ; ax=204h -> Get PM interrupt vector
        int 31h
        pop eax

@@exit:
        popfd
        ret

;
; Set interrupt vector
; In:
;   ds = data32sel
;   bl = interrupt number
;   cx:edx = selector:offset of exception handler
;   If IDT entries 8-15 are shared between exceptions and IRQs:
;     CF=0 -> Set IRQ handler
;     CF=1 -> Set exception handler
; Out:
;   CF=0 -> Interrupt vector successfully set
;   CF=1 -> Interrupt number is not present in the IDT (int number too high)
;
setvect:
        pushfd
        cli
        cmp systemtype,3                ; Under DPMI the IDT is not part of
        je @@DPMIsetvect                ;  this program in memory

        push eax                        ; Check if int is in the IDT
        mov ax,idtr                     ; idtr[0] contains the IDT limit
        shr ax,3                        ; An int gate is 8 bytes in size
        cmp bl,al                       ; ax < 256, because highest int < 256
        pop eax                         ; Highest int that is allowed is
        ja @@errexit                    ;  IDT limit/8, not IDT size/8

        push ebx
        and ebx,0ffh
        cmp cs:mPICbase,8               ; Check if IDT entries 8-15 are
        jne short @@notshared           ;  shared
        cmp ebx,8                       ; Check if int is one of the shared
        jb short @@notshared            ;  IDT entries
        cmp ebx,15
        ja short @@notshared

        test byte ptr [esp+4],1         ; Test CF in flags on stack
        jnz short @@setexc              ; CF set -> Set exception handler
        sub ebx,8
        imul ebx,IRQ1handler-IRQ0handler
        mov IRQhandler[ebx+4],cx
        mov dword ptr IRQhandler[ebx],edx
        pop ebx
        jmp short @@exit
@@setexc:
        sub ebx,8
        imul ebx,exc9handler-exc8handler
        mov exchandler[ebx+4],cx
        mov dword ptr exchandler[ebx],edx
        pop ebx
        jmp short @@exit

@@notshared:
        push edi edx
        mov edi,idt
        shl ebx,3                       ; ebx=offset of int gate dsc in IDT
        add edi,ebx
        mov [edi+2],cx                  ; int gate[2]=selector
        mov [edi],dx                    ; int gate[0]=LSW of offset
        shr edx,16
        mov [edi+6],dx                  ; int gate[6]=MSW of offset
        pop edx edi ebx
        jmp short @@exit

@@DPMIsetvect:
        push eax
        mov ax,205h                     ; ax=205h -> Set PM interrupt vector
        int 31h
        pop eax
        jnc short @@exit

@@errexit:
        popfd
        stc
        ret
@@exit:
        popfd
        clc
        ret

;
; Map a region of physical addresses to linear addresses
; In:
;   ds = data32sel
;   eax = size of region to map in bytes
;   edx = physical address of memory
; Out:
;   CF=0 -> Memory has been successfully mapped
;     edx = linear address that corresponds with physical memory
;   CF=1 -> Not enough memory to create page tables / Page directory is full
;     edx = ?
;
map_phys_addr:
        push eax ebx ecx esi edi es
        cmp systemtype,3                ; Under DPMI the page tables are not
        je @@DPMImap                    ;  part of this program in memory
        cmp systemtype,2
        je short @@VCPImap

        clc                             ; Under raw and XMS paging is not
        jmp @@exit                      ;  enabled -> lin addr=phys addr

@@VCPImap:
        add eax,0fffh                   ; Round region up to 4K units
        and eax,0fffff000h
        shr eax,10                      ; eax=amount mem needed to map mem;
        mov ecx,eax                     ;  for every KB 1 byte in page table

        mov ebx,himembase
        add ebx,code32a
        mov esi,ebx
        add esi,0fffh                   ; Check free memory base alignment
        and esi,0fffff000h
        sub esi,ebx                     ; esi=offset to get at 4K alignment
        add eax,esi                     ;  from current free high mem base
                                        ; Extra mem in eax for aligning ->
                                        ;  it is allowed to do aligning
        call gethimem
        jc @@errexit
        add eax,code32a                 ; eax=linear address of 1st page
        add eax,0fffh                   ;  table entry
        and eax,0fffff000h              ; Align address on 4KB byte boundary
        mov edi,eax                     ; edi=lin address of 1st table entry
        push ecx edi                    ; Save ecx and edi for making page
                                        ;  table entries
        mov es,zerosel
        mov edi,VCPIdir_base
        mov ecx,4096/4
        xor eax,eax                     ; Search for first 0 (=free) entry
        cld
        repne scasd                     ; Compare eax with es:edi while not 0
        jne @@errexit                   ; ZF=0 (ecx=0) -> No page dir entries 
                                        ;  free
        sub edi,4                       ; edi=lin address of 1st free entry
        sub edi,VCPIdir_base
        shl edi,20                      ; edi=lin address of mapped phys mem
        mov eax,edi

        pop edi ecx                     ; Values of ecx and edi are needed
        push ecx edi                    ;  again
        and edx,0fffff000h              ; Clear page attributes
        or edx,7                        ; User (=page addressable regardless
        add ecx,edi                     ;  the CPL); Read and write; Present
@@next_page:                            ; Amount mem that page tables are 
        mov es:[edi],edx                ;  created for is diff ecx and edi
        add edx,4096                    ; Next page resides 4K higher in mem
        add edi,4
        cmp edi,ecx
        jb short @@next_page
        mov edx,eax                     ; edx=lin address of mapped phys mem

        pop ebx                         ; ebx=lin address of 1st table entry
        shr ebx,12                      ; ebx=number of 4KB page where in mem
        mov esi,VCPIdir_base
        mov edi,edx                     ; edi=lin address of mapped phys mem
        shr edi,20                      ; edi=offset of entry of 1st page
        add edi,VCPIdir_base            ;  table in page dir (=1MB units mem)
        pop ecx                         ; ecx=size of region to map in KB
        add ecx,0fffh                   ; Round up to 4MB units
        shr ecx,12                      ; ecx=number of 4MB regions to map
@@next_table:
        mov eax,es:[ebx*4 + esi+4096]
        and eax,0fffff000h              ; Clear page attributes
        or eax,7                        ; User (=page addressable regardless
        stosd                           ;  the CPL); Read and write; Present
        inc ebx                         ; For each 4MB 4 bytes in directory
        loop @@next_table               ;  -> for each MB 1 byte

        clc
        jmp short @@exit

@@DPMImap:
        mov cx,dx
        shld ebx,edx,16                 ; bx:cx=physical address of memory
        mov di,ax
        shld esi,eax,16                 ; si:di=size of region to map (bytes)
        mov ax,800h                     ; ax=800h -> Map physical address to
        int 31h                         ;  linear address
        jc short @@errexit
        shrd edx,ebx,16                 ; bx:cx=lin addr to access phys mem
        mov dx,cx
        clc                             ; CF set only on error
        jmp short @@exit

@@errexit:
        stc
@@exit:
        pop es edi esi ecx ebx eax
        ret

;
; Shell to DOS
; In:
;   ds = data32sel
;   esi = offset of 0 terminated string (program to start via COMMAND.COM)
;   esi = 0 -> Only COMMAND.COM will be started
; Out:
;   CF=0 -> No error
;     lomembase = lomembase + 16 and aligned on paragraph boundary
;   CF=1 -> Error
;
; Note:
;   int 21h/ah=4ah places 1 byte + 2 words at first paragraph above lomembase
;   if lomembase is not already on paragraph boundary. These bytes are part
;   of the 16 byte Memory Control Block and must not be overwritten. dosshell
;   saves this MCB ("add lomembase,16").
;
; Offsets in Program Segment Prefix:
; 5Ch -> default unopened FCB #1
; 6Ch -> default unopened FCB #2
; 80h -> count of bytes in command tail (inluding 1st space, not last cr)
; 81h -> all characters entered after the prog name, followed by a cr

; Parameter Block:
; dw      env_seg       ; segment of environment block
; dd      tail          ; segment:offset of command tail (counter)
; dd      fcb1          ; segment:offset of default FCB #1
; dd      fcb2          ; segment:offset of default FCB #2
;
dosshell:
        push eax ecx esi edi es
        cld
                                        ; Make sure allocated low mem won't
        mov ax,psp_seg                  ;  be overwritten by DOS
        mov v86r_es,ax                  ; es=segment of the block
        mov eax,code32a
        add eax,lomembase               ; eax=offset 1st byte in free low mem
        add eax,0fh                     

        mov ebx,eax
        and ebx,0fffffff0h              ; eax=idem, rounded up to paragraphs
        sub ebx,code32a                 ; Align lomembase on paragraph
        mov lomembase,ebx               ;  boundary -> MCB at lomembase

        shr eax,4                       ; ax=offset in paragraph units
        sub ax,v86r_es                  ; ax=new block size in paragraphs
        mov v86r_bx,ax                  ; bx=new block size in paragraphs
        mov v86r_ah,4ah                 ; ah=4ah -> Modify allocated mem block
        mov al,21h
        int RMCALL_VECT
        test v86r_flags,1               ; Error?
;       jnz @@errexit                   ; ?: Under CWSDPMI ah=4ah fails
        add lomembase,16                ; Save MCB

        cmp esi,0                       ; Start only COMMAND.COM?
        jne short @@startprog

        mov cmdtail[0],0                ; No arguments for COMMAND.COM
                                        ; Copy segment addresses of:
        mov cx,env_seg                  ;  environment (setting paramblock[0]
        mov paramblock[0],cx            ;  to 0 also indicates that EXEC func
        mov ax,psp_seg                  ;  must use copy of current env), 
        mov paramblock[8],ax            ;  default unopened FCB #1 and
        mov paramblock[12],ax           ;  default unopened FCB #2 of this
                                        ;  program to param block of 'child'.
        jmp short @@get_comspec
@@startprog:                            ; Start program at ds:esi
        mov ax,ds
        mov es,ax                       ; es=ds for scasb
        mov edi,esi                     ; Get string length of program name
        mov ecx,0ffffffffh              ; Search max 4GB
        mov al,0                        ; Scan for 0 (=end of string)
        repne scasb                     ; Compare al with es:edi
        not ecx                         ; When scasb stops, edi points 1 byte
        dec ecx                         ;  past the 0; that's 1 byte too far
                                        ; ecx=length of program name

        mov edi,offset cmdtail+1        ; Tail starts 1 byte past 'cmdtail'
        add edi,presettail              ; Save preset bytes in 'cmdtail'
        rep movsb                       ; Copy program's name to tail

        sub edi,offset cmdtail+1        ; edi=length of command tail (Don't
        mov ax,di                       ;  count command tail counter)
        mov cmdtail[0],al               ; Length cannot be larger than 127

        mov cx,env_seg
        mov paramblock[0],cx
        mov ax,psp_seg
        mov paramblock[8],ax
        mov paramblock[12],ax

@@get_comspec:                          ; Get COMSPEC environment variable
        mov es,envsel
        mov al,0                        ; Scan for 0 (=end of string)
        xor edi,edi

@@search_comspec:
        mov esi,offset comspec
        cmpsd                           ; Compare ds:esi with es:edi; 'COMS'
        jne short @@end_of_string
        cmpsd                           ; Compare ds:esi with es:edi; 'PEC='
        je short @@exec
        jmp short @@errexit

@@end_of_string:
        scasb                           ; Compare es:edi with al; scan until
        jne short @@end_of_string       ;  end of environment string is found
        jmp short @@search_comspec

@@exec:
        mov v86r_ax,4b00h               ; ax=4b00h -> EXEC/Load and execute
        mov v86r_bx,small offset paramblock ; program 
        mov v86r_dx,di
        mov v86r_ds,cx                  ; ds:dx=pointer to ASCIIz filename
        mov v86r_es,code32              ; es:bx=pointer to parameter block
        mov al,21h
        int RMCALL_VECT
        bt v86r_flags,0                 ; Copy virtual CF value to CF
        jmp short @@exit
@@errexit:
        stc
@@exit:
        pop es edi esi ecx eax
        ret

;
; Print '$' terminated string via DOS
; In:
;   ds = data32sel
;   edx = offset of '$' terminated string (string must be in low memory)
;
dosprint:
        push eax edx
        add edx,code32a
        mov al,dl
        and eax,0fh
        shr edx,4
        mov v86r_ds,dx
        mov v86r_dx,ax
        mov v86r_ah,9                   ; ah=9 -> Print string at ds:dx
        mov al,21h
        int RMCALL_VECT
        pop edx eax
        ret

;
; Put 0 terminated string to es:edi
; In:
;   ds:esi = address of 0 terminated string
;   bl = character attribute
; Out:
;   edi = 1 byte past last written character (attribute)
;
putstr:
        push ecx esi eax
        push edi es
        mov ax,ds                       ; First, get string length
        mov es,ax                       ; es=ds for scasb
        mov edi,esi
        mov ecx,0ffffffffh              ; Search max 4GB
        mov al,0                        ; Scan for 0 (=end of string)
        cld
        repne scasb                     ; Compare al with es:edi
        not ecx
        dec ecx                         ; When scasb stops, edi points 1 byte
                                        ;  past the 0; that's 1 byte too far
        pop es edi                      ; ecx=string length
        mov ah,bl                       ; bl=character attribute
@@putchar:
        lodsb
        stosw
        loop @@putchar
        pop eax esi ecx
        ret

code32  ends

code32end segment para public
code32end ends

        end     start16
