; This program is made by Daniel Horchner.
; email: Daniel.Horchner@95.Student.WAU.NL
;        dbjh@usa.net
;
; This example runs V86 mode code at IOPL 3. So, STI isn't detected, but IRQs
; must not be enabled -> disable IRQs by setting the PIC masks.

        .386p
        locals

code32  segment para public use32
        assume cs:code32, ds:code32, ss:code32

include raw32.inc

public  main

;32-bit data
TSS1dsc         seg_descriptor  <2067h,,, 89h, 0,>
TSS1            TSS     <>
                db      2000h dup(0)    ; 1 bit for each port; 64K / 8 = 8K
TSS1sel         dw      0

org_exc13       dd      0               ; offset and selector of original
                dw      0               ;  exception 13 handler
org_tr          dw      0               ; original Task Register contents
mPICmask        db      0               ; original master PIC mask
sPICmask        db      0               ; original slave PIC mask

msg0            db      'Not running at Privilege Level 0.','$'
msg1            db      'Message from exception 13 handler.',0
msg2            db      'Press a key to continue.',0

;32-bit code
main:
                                        ; First, check if running at PL 0
        mov ebx,cs
        lar ecx,ebx
        and ecx,6000h                   ; Only the DPL bits are needed
        jz short @@PL0
        mov edx,offset msg0
        call dosprint
        jmp exit
@@PL0:
;
        mov eax,codeend
        shl eax,4
        sub eax,code32a
        call getlomem                   ; Save V86 code segment

        str org_tr                      ; Save original Task Register

        mov cx,1                        ; Allocate 1 descriptor for TSS
        call getdsc
        jc @exit
        mov TSS1sel,ax

        mov eax,code32a
        add eax,offset TSS1
        mov TSS1dsc.base0_15,ax
        shr eax,16
        mov TSS1dsc.base16_23,al
        mov TSS1dsc.base24_31,ah

        mov es,data32sel
        mov edi,offset TSS1dsc
        mov bx,TSS1sel                  ; bx=selector
        call setdsc
        jc @exit

        cli
        ltr TSS1sel
;
        mov bl,13
        stc
        call getvect
        mov word ptr org_exc13[4],cx    ; cx:edx=addr of exception handler
        mov org_exc13[0],edx

        mov cx,cs                       ; Install V86 monitor
        mov edx,offset new_exc13
        stc
        call setvect
        jc @exit

        in al,21h                       ; Save PIC masks
        mov mPICmask,al
        in al,0a1h
        mov sPICmask,al
        mov al,0ffh                     ; Mask all IRQs
        out 21h,al
        out 0a1h,al

        push ds es fs gs
        mov TSS1.ss0,ss                 ; Save protected mode ss:esp in Task
        mov TSS1.esp0,esp               ;  State Segment -> V86 monitor uses 
                                        ;  current stack
                                        ; Build stack frame for entering V86 
        push dword ptr v86r_gs          ;  mode: gs,
        push dword ptr v86r_fs          ;  fs,
        push dword ptr v86r_ds          ;  ds,
        push dword ptr v86r_es          ;  es,
        push large v86code              ;  ss,
        push large offset v86stackbase  ;  esp,
        pushfd                          ;  eflags,
        or [esp],20000h                 ; Set virtual 8086 mode flag
        push large v86code              ;  cs,
        push large offset v86start      ;  eip
        iretd                           ; VM bit in eflags on stack is set ->
retfromV86:                             ;  get eip, cs, eflags, esp, ss, es,
        pop gs fs es ds                 ;  ds, fs, gs from stack -> V86 mode

        mov bl,13
        mov cx,word ptr org_exc13[4]    ; cx:edx=addr of exception handler
        mov edx,org_exc13[0]
        stc
        call setvect

        mov al,mPICmask                 ; Restore original PIC masks
        out 21h,al
        mov al,sPICmask
        out 0a1h,al

@exit:
        sub esp,6
        sgdt [esp]
        mov eax,[esp+2]
        add esp,6
        movzx ebx,org_tr
        and ebx,not 3                   ; ebx=offset of TSS dsc in GDT
        and byte ptr gs:[eax+ebx+5],not 2 ; Clear Busy bit
        ltr org_tr                      ; Restore original Task Register
        sti

        mov esi,offset msg2
        @rlp edi,0b8000h+2*160
        mov bl,1fh
        call putstr
        mov v86r_ah,0                   ; ah=0 -> Wait for key and read char
        mov al,16h
        int RMCALL_VECT
        jmp exit                        ; Return to real/V86 mode

;
new_exc13:                              ; General protection
        test byte ptr [esp+14],2        ; exc in V86 -> gs,fs,ds,es,ss,esp,
        jnz short V86monitor            ;  eflags,cs,eip,error code on stack
        jmp fword ptr org_exc13         ; Jump to 'normal' handler
V86monitor:                             ; Must remove error code from stack
        add esp,4                       ;  before IRETD instruction
        pushad                          ; eax,ecx,edx,ebx,esp,ebp,esi,edi =
        mov ds,cs:zerosel               ;  32 bytes pushed on stack
        movzx ebx,word ptr [esp+36]     ; ebx=V86 cs
        shl ebx,4                       ; ebx=linear address of cs
        add ebx,[esp+32]                ; ebx=linear address of instruction
        inc word ptr [esp+32]           ;  that caused exception
        mov al,[ebx]                    ; eip on stack points 1 byte past 
        mov dl,3                        ;  instruction that caused exception
        cmp al,0cch                     ; Is it a 1 byte INT 3 instruction?
        je short emu_int
        mov dl,4
        cmp al,0ceh                     ; Is it a 1 byte INTO instruction?
        je short emu_int
        cmp al,0cdh                     ; Is it a 2 byte INT instruction?
        je short @@INTopcode
        popad
        int 15                          ; Indicate unhandled error while V86
@@INTopcode:
        inc word ptr [esp+32]           ; eip on stack 1 byte past int number
        mov dl,[ebx+1]                  ; dl=int number to call
        cmp dl,RMCALL_VECT              ; Exit V86 mode?
        jne short emu_int
        add esp,68                      ; Remove 9+8 dwords from the stack ->
        jmp retfromV86                  ;  original PM esp
emu_int:                                ; Emulate INT instruction
        mov ds,cs:zerosel
        movzx ebx,dl
        shl ebx,2                       ; ebx=address of real mode int vector
        movzx edx,word ptr [esp+48]     ; edx=V86 ss
        shl edx,4                       ; edx=linear address of ss 
        sub word ptr [esp+44],6         ; Reserve 6 bytes on V86 stack for
        add edx,[esp+44]                ;  flags + return address after int 
        mov ax,[esp+40]                 ; edx=linear address top of V86 stack
        mov [edx+4],ax                  ; ax=V86 flags
        mov ax,[esp+36]                 ; ax=V86 cs
        mov [edx+2],ax
        mov ax,[esp+32]                 ; ax=V86 ip
        mov [edx],ax                    ; Return to V86_int after int
        and byte ptr [esp+41],not 3     ; Clear IF and TF (trap flag)
        mov eax,[ebx]
        mov [esp+32],ax                 ; Put ip of int handler on stack
        shr eax,16
        mov [esp+36],ax                 ; Put cs of int handler on stack
        popad                           ; VM bit in eflags on stack is set ->
        iretd                           ;  after iretd -> back in V86 mode

code32  ends

v86code segment para public use16
        assume cs:v86code, ds:v86code, ss:v86code

;16-bit data
                db      256*8 dup(0)    ; ? -> Read RAW32.TXT!
v86stack        db      100 dup(0)      ; V86 mode stack
v86stackbase    label

v86msg          db      'Message from Virtual 8086 mode.'
v86msglen       =       $-v86msg
v86msg2         db      'V86 message printed by emulated int 10h.'
v86msg2len      =       $-v86msg2

;16-bit code
v86start:
        mov ax,cs
        mov ds,ax
        mov si,offset v86msg
        mov ax,0b800h
        mov es,ax
        mov di,0
        mov cx,v86msglen
        mov ah,1fh
@@putchar:                              ; Print first message
        lodsb
        stosw
        loop @@putchar

        mov ah,13h                      ; ah=13h -> Write string
        mov al,0                        ; ah=0 -> str chars only, attr in bl
        mov bh,0                        ; bh=video page number
        mov bl,1fh                      ; bl=attribute if al bit 1=0
        mov cx,v86msg2len               ; cx=length of string
        mov dh,1                        ; dh=row coordinate
        mov dl,0                        ; dl=column coordinate
        mov di,seg v86msg2
        mov es,di
        mov bp,offset v86msg2           ; es:bp=pointer to string
        int 10h                         ; Emulate INT instruction

        int RMCALL_VECT                 ; int with DPL < 3 or not in IDT ->
                                        ;  exc 13. V86 monitor checks for int
                                        ;  RMCALL_VECT

v86code ends

codeend segment para public
codeend ends

        end
